# This file is obsolete - it was created by special request to allow limited import from IBKR HTML reports
# Import interface was changed so it was removed from menu but kept as a reference

import logging
import pandas
from datetime import datetime, timezone
from PySide6.QtWidgets import QApplication
from jal.db.db import JalDB

class IBKR_obsolete:
    def __init__(self, filename):
        self._filename = filename
        self._account = ''

    def tr(self, text):
        return QApplication.translate("IBKR", text)

    def load(self):
        # Read first table with account details
        try:
            data = pandas.read_html(self._filename, match='Account Capabilities')
        except:
            logging.error(self.tr("Can't read statement file"))
            return False
        if len(data) != 1:
            logging.error(self.tr("Can't get account details from the statement"))
            return False
        for i, row in data[0].iterrows():
            if row[0] == 'Account':
                self._account = row[1]
        if self._account == '':
            logging.error(self.tr("Can't get account number from the statement"))
            return False

        # Read Trades table
        try:
            data = pandas.read_html(self._filename, match='Date/Time', attrs = {'id': 'summaryDetailTable'})
        except:
            logging.error(self.tr("Can't read Trades table from statement file"))
            return False
        if len(data) != 1:
            logging.error(self.tr("Can't get Trades table from the statement"))
            return False
        statement = data[0]
        statement = statement[statement['Symbol'].notna()]

        account_id = None
        for i, row in statement.iterrows():
            if row[0] == 'Forex':  # We reached end of Stock trades
                break
            if row[0].startswith('Total') or row[0] == 'Stocks' or row[0] == 'Symbol':  # Skip totals and headers
                continue
            if row[0] == row[1]:  # it's a currency header - match account
                account_id = JalDB().get_account_id(self._account, row[0])
                continue
            if account_id is None:  # no reason to check further if we don't have valid account here
                continue
            asset_id = JalDB().get_asset_id(row[0])
            if asset_id is None:
                logging.warning(self.tr("Unknown asset ") + f"'{row[0]}'")
                continue
            timestamp = int(datetime.strptime(row[1], "%Y-%m-%d, %H:%M:%S").replace(tzinfo=timezone.utc).timestamp())
            settlement = int(datetime.strptime(row[1][:10], "%Y-%m-%d").replace(tzinfo=timezone.utc).timestamp())
            number = f"{i}"
            qty = float(row[2])
            price = float(row[3])
            fee = float(row[6])
            JalDB().add_trade(account_id, asset_id, timestamp, settlement, number, qty, price, -fee)
        logging.info(self.tr("Load IBKR Activity statement completed"))
        return True



from functools import partial

from PySide6.QtCore import Qt, Slot, Signal, QObject, QDateTime
from PySide6.QtGui import QAction
from PySide6.QtWidgets import QMenu
from jal.ui.reports.ui_holdings_report import Ui_HoldingsWidget
from jal.db.settings import JalSettings
from jal.db.holdings_model import HoldingsModel
from jal.widgets.mdi import MdiWidget
from jal.db.tax_estimator import TaxEstimator
from jal.widgets.price_chart import ChartWindow

JAL_REPORT_CLASS = "HoldingsReport"


# ----------------------------------------------------------------------------------------------------------------------
class HoldingsReport(QObject):
    def __init__(self):
        super().__init__()
        self.name = self.name = self.tr("Holdings")
        self.window_class = "HoldingsReportWindow"


# ----------------------------------------------------------------------------------------------------------------------
class HoldingsReportWindow(MdiWidget, Ui_HoldingsWidget):
    def __init__(self, parent=None):
        MdiWidget.__init__(self, parent)
        self.setupUi(self)
        self.parent_mdi = parent

        self.holdings_model = HoldingsModel(self.HoldingsTableView)
        self.HoldingsTableView.setModel(self.holdings_model)
        self.holdings_model.configureView()
        self.HoldingsTableView.setContextMenuPolicy(Qt.CustomContextMenu)

        self.connect_signals_and_slots()

        # Setup holdings parameters
        current_time = QDateTime.currentDateTime()
        current_time.setTimeSpec(Qt.UTC)  # We use UTC everywhere so need to force TZ info
        self.HoldingsDate.setDateTime(current_time)
        self.HoldingsCurrencyCombo.setIndex(JalSettings().getValue('BaseCurrency'))

    def connect_signals_and_slots(self):
        self.HoldingsDate.dateChanged.connect(self.HoldingsTableView.model().setDate)
        self.HoldingsCurrencyCombo.changed.connect(self.HoldingsTableView.model().setCurrency)
        self.HoldingsTableView.customContextMenuRequested.connect(self.onHoldingsContextMenu)

    @Slot()
    def onHoldingsContextMenu(self, pos):
        index = self.HoldingsTableView.indexAt(pos)
        contextMenu = QMenu(self.HoldingsTableView)
        actionShowChart = QAction(text=self.tr("Show Price Chart"), parent=self.HoldingsTableView)
        actionShowChart.triggered.connect(partial(self.showPriceChart, index))
        contextMenu.addAction(actionShowChart)
        actionEstimateTax = QAction(text=self.tr("Estimate Russian Tax"), parent=self.HoldingsTableView)
        actionEstimateTax.triggered.connect(partial(self.estimateRussianTax, index))
        contextMenu.addAction(actionEstimateTax)
        contextMenu.popup(self.HoldingsTableView.viewport().mapToGlobal(pos))

    @Slot()
    def showPriceChart(self, index):
        model = index.model()
        account, asset, asset_qty = model.get_data_for_tax(index)
        self.parent_mdi.addSubWindow(ChartWindow(account, asset, asset_qty))

    @Slot()
    def estimateRussianTax(self, index):
        model = index.model()
        account, asset, asset_qty = model.get_data_for_tax(index)
        self.parent_mdi.addSubWindow(TaxEstimator(account, asset, asset_qty))



import logging
from datetime import datetime, timezone, timedelta

from jal.constants import Setup, PredefinedCategory
from jal.data_import.statement import FOF, Statement_ImportError
from jal.data_import.statement_xls import StatementXLS

JAL_STATEMENT_CLASS = "StatementKIT"


# ----------------------------------------------------------------------------------------------------------------------
class StatementKIT(StatementXLS):
    Header = (4, 0, "КИТ Финанс (АО)")
    PeriodPattern = (5, 8, r"(?P<S>\d\d\.\d\d\.\d\d\d\d)\s.\s(?P<E>\d\d\.\d\d\.\d\d\d\d)")
    AccountPattern = (5, 5, r"(?P<ACCOUNT>.*)-(.*)")
    SummaryHeader = "Состояние денежных средств на счете"

    asset_section = "Состояние портфеля ценных бумаг"
    asset_columns = {
        "name": "Наименование\nЦБ ",
        "isin": "ISIN",
        "reg_number": "Код гос. регистрации"
    }

    def __init__(self):
        super().__init__()
        self.name = self.tr("KIT Finance")
        self.icon_name = "kit.png"
        self.filename_filter = self.tr("KIT Finance statement (*.xlsx)")

    def _load_deals(self):
        cnt = 0
        columns = {
            "number": "Номер\nсделки",
            "date": "Дата сделки",
            "time": "Время сделки",
            "asset": "Наименование\nЦБ",
            "isin": "ISIN",
            "B/S": "Тип операции",
            "price": "Цена сделки ",
            "qty": "Количество",
            "amount": "Сумма сделки",
            "currency": " Валюта\nзаключения\nсделки",
            "accrued_int": " НКД",
            "settlement": "Дата поставки\n\(план.\)",
            "fee_ex": "Комиссия\nТС",
            "fee_broker": "Комиссия\nброкера"
        }

        row, headers = self.find_section_start("Заключенные сделки с ценными бумагами", columns)
        if row < 0:
            return False
        while row < self._statement.shape[0]:
            if self._statement[0][row] == '' and self._statement[0][row + 1] == '':
                break
            try:
                code = self.currency_id(self.currency_substitutions[self._statement[headers['currency']][row]])
            except KeyError:
                code = self.currency_id(self._statement[headers['currency']][row])
            asset_id = self.asset_id({'isin': self._statement[headers['isin']][row],
                                      'currency': code, 'search_online': "MOEX"})
            if self._statement[headers['B/S']][row] == 'Покупка':
                amount = -self._statement[headers['amount']][row]
                qty = self._statement[headers['qty']][row]
                bond_interest = -self._statement[headers['accrued_int']][row]
            elif self._statement[headers['B/S']][row] == 'Продажа':
                amount = self._statement[headers['amount']][row]
                qty = self._statement[headers['qty']][row]
                bond_interest = self._statement[headers['accrued_int']][row]
            else:
                row += 1
                logging.warning(self.tr("Unknown trade type: ") + self._statement[headers['B/S']][row])
                continue
            price = self._statement[headers['price']][row]
            fee = round(abs(self._statement[headers['fee_ex']][row] + self._statement[headers['fee_broker']][row]), 8)
            if abs(abs(price * qty) - amount) >= Setup.DISP_TOLERANCE:
                price = abs(amount / qty)
            number = self._statement[headers['number']][row]
            # Dates are loaded as datetime objects but time is loaded as string
            t_date = self._statement[headers['date']][row]
            t_time = datetime.strptime(self._statement[headers['time']][row], "%H:%M:%S").time()
            trade_datetime = t_date + timedelta(hours=t_time.hour, minutes=t_time.minute, seconds=t_time.second)
            timestamp = int(trade_datetime.replace(tzinfo=timezone.utc).timestamp())
            settlement = int(self._statement[headers['settlement']][row].replace(tzinfo=timezone.utc).timestamp())
            account_id = self._find_account_id(self._account_number, self._statement[headers['currency']][row])
            new_id = max([0] + [x['id'] for x in self._data[FOF.TRADES]]) + 1
            trade = {"id": new_id, "number": str(number), "timestamp": timestamp, "settlement": settlement,
                     "account": account_id, "asset": asset_id, "quantity": qty, "price": price, "fee": fee}
            self._data[FOF.TRADES].append(trade)
            if bond_interest != 0:
                new_id = max([0] + [x['id'] for x in self._data[FOF.ASSET_PAYMENTS]]) + 1
                payment = {"id": new_id, "type": FOF.PAYMENT_INTEREST, "account": account_id, "timestamp": timestamp,
                           "number": str(number), "asset": asset_id, "amount": bond_interest, "description": "НКД"}
                self._data[FOF.ASSET_PAYMENTS].append(payment)
            cnt += 1
            row += 1
        logging.info(self.tr("Trades loaded: ") + f"{cnt}")

    def _load_cash_transactions(self):
        cnt = 0
        columns = {
            "date": "Дата",
            "operation": "Тип операции",
            "amount": "Сумма",
            "currency": " Валюта",
            "reason": "Основание",
            "note": "Примечание",
            "market": "Секция"
        }
        operations = {
            'Внесение д/с в торг': self.transfer_in,
            'Вывод дс': self.transfer_out,
            'Ком бр аб плата спот': self.fee,
            'Комиссия НРД': self.fee,
            'Delta-long': self.fee,
            '% по займу ЦБ': self.interest,
            'Налог с див.доход ФЛ': self.tax
        }
        row, headers = self.find_section_start("Движение денежных средств по неторговым операциям", columns)
        if row < 0:
            return False
        while row < self._statement.shape[0]:
            if self._statement[0][row] == '':
                break
            operation = self._statement[headers['operation']][row]
            if operation not in operations:  # not supported type of operation
                raise Statement_ImportError(self.tr("Unsuppported cash transaction ") + f"'{operation}'")
            timestamp = int(self._statement[headers['date']][row].replace(tzinfo=timezone.utc).timestamp())
            account_id = self._find_account_id(self._account_number, self._statement[headers['currency']][row])
            amount = self._statement[headers['amount']][row]
            reason = self._statement[headers['reason']][row]
            description = self._statement[headers['note']][row]
            operations[operation](timestamp, account_id, amount, reason, description)
            cnt += 1
            row += 1
        logging.info(self.tr("Cash operations loaded: ") + f"{cnt}")

    def transfer_in(self, timestamp, account_id, amount, reason, note):
        account = [x for x in self._data[FOF.ACCOUNTS] if x["id"] == account_id][0]
        description = reason + ", " + note
        new_id = max([0] + [x['id'] for x in self._data[FOF.TRANSFERS]]) + 1
        transfer = {"id": new_id, "account": [0, account_id, 0],
                    "asset": [account['currency'], account['currency']], "timestamp": timestamp,
                    "withdrawal": amount, "deposit": amount, "fee": 0.0, "description": description}
        self._data[FOF.TRANSFERS].append(transfer)

    def transfer_out(self, timestamp, account_id, amount, reason, note):
        account = [x for x in self._data[FOF.ACCOUNTS] if x["id"] == account_id][0]
        description = reason + ", " + note  # amount is negative in XLSX file
        new_id = max([0] + [x['id'] for x in self._data[FOF.TRANSFERS]]) + 1
        transfer = {"id": new_id, "account": [account_id, 0, 0],
                    "asset": [account['currency'], account['currency']], "timestamp": timestamp,
                    "withdrawal": -amount, "deposit": -amount, "fee": 0.0, "description": description}
        self._data[FOF.TRANSFERS].append(transfer)

    def fee(self, timestamp, account_id, amount, _reason, description):
        new_id = max([0] + [x['id'] for x in self._data[FOF.INCOME_SPENDING]]) + 1
        fee = {"id": new_id, "timestamp": timestamp, "account": account_id, "peer": 0,
               "lines": [{"amount": amount, "category": -PredefinedCategory.Fees, "description": description}]}
        self._data[FOF.INCOME_SPENDING].append(fee)

    def interest(self, timestamp, account_id, amount, _reason, description):
        new_id = max([0] + [x['id'] for x in self._data[FOF.INCOME_SPENDING]]) + 1
        interest = {"id": new_id, "timestamp": timestamp, "account": account_id, "peer": 0,
                    "lines": [{"amount": amount, "category": -PredefinedCategory.Interest, "description": description}]}
        self._data[FOF.INCOME_SPENDING].append(interest)

    def tax(self, timestamp, account_id, amount, _reason, description):
        logging.info(self.tr("Dividend taxes are not supported for KIT broker statements yet"))



from PySide6.QtCore import Qt, Slot, QObject
from PySide6.QtSql import QSqlTableModel
from jal.ui.reports.ui_deals_report import Ui_DealsReportWidget
from jal.db.helpers import db_connection, executeSQL
from jal.db.operations import CorporateAction
from jal.widgets.delegates import TimestampDelegate, FloatDelegate
from jal.widgets.mdi import MdiWidget

JAL_REPORT_CLASS = "DealsReport"


# -----------------------------------------------------------------------------------------------------------------------
class DealsReportModel(QSqlTableModel):
    def __init__(self, parent_view):
        self._columns = [("asset", self.tr("Asset")),
                         ("o_datetime", self.tr("Open Date")),
                         ("c_datetime", self.tr("Close Date")),
                         ("open_price", self.tr("Open Price")),
                         ("close_price", self.tr("Close Price")),
                         ("qty", self.tr("Qty")),
                         ("fee", self.tr("Fee")),
                         ("profit", self.tr("P/L")),
                         ("rel_profit", self.tr("P/L, %")),
                         ("corp_action", self.tr("Note"))]
        self.ca_names = {
            CorporateAction.SymbolChange: self.tr("Symbol change"),
            CorporateAction.Split: self.tr("Split"),
            CorporateAction.SpinOff: self.tr("Spin-off"),
            CorporateAction.Merger: self.tr("Merger")
        }
        self._view = parent_view
        self._begin = 0
        self._end = 0
        self._account_id = 0
        self._group_dates = 0
        self._query = None
        self._timestamp_delegate = None
        self._float_delegate = None
        self._float2_delegate = None
        self._float4_delegate = None
        self._profit_delegate = None
        self._ca_delegate = None
        QSqlTableModel.__init__(self, parent=parent_view, db=db_connection())

    def setColumnNames(self):
        for column in self._columns:
            self.setHeaderData(self.fieldIndex(column[0]), Qt.Horizontal, column[1])

    def resetDelegates(self):
        for column in self._columns:
            self._view.setItemDelegateForColumn(self.fieldIndex(column[0]), None)

    def data(self, index, role=Qt.DisplayRole):
        if role == Qt.DisplayRole:
            if index.column() == self.fieldIndex("corp_action"):
                try:
                    ca_type = int(super().data(index, role))
                except ValueError:
                    ca_type = 0
                if ca_type > 0:
                    text = self.tr(" Opened with ") + self.ca_names[ca_type]
                elif ca_type < 0:
                    text = self.tr(" Closed with ") + self.ca_names[-ca_type]
                else:
                    text = ''
                return text
        return super().data(index, role)

    def configureView(self):
        self._view.setModel(self)
        self.setColumnNames()
        self.resetDelegates()
        font = self._view.horizontalHeader().font()
        font.setBold(True)
        self._view.horizontalHeader().setFont(font)
        self._view.setColumnWidth(self.fieldIndex("asset"), 300)
        self._view.setColumnWidth(self.fieldIndex("corp_action"), 200)
        self._view.setColumnWidth(self.fieldIndex("o_datetime"),
                                  self._view.fontMetrics().horizontalAdvance("00/00/0000 00:00:00") * 1.1)
        self._view.setColumnWidth(self.fieldIndex("c_datetime"),
                                  self._view.fontMetrics().horizontalAdvance("00/00/0000 00:00:00") * 1.1)
        if self._group_dates == 1:
            self._timestamp_delegate = TimestampDelegate(display_format='%d/%m/%Y')
        else:
            self._timestamp_delegate = TimestampDelegate()
        self._view.setItemDelegateForColumn(self.fieldIndex("o_datetime"), self._timestamp_delegate)
        self._view.setItemDelegateForColumn(self.fieldIndex("c_datetime"), self._timestamp_delegate)
        self._float_delegate = FloatDelegate(0, allow_tail=True)
        self._view.setItemDelegateForColumn(self.fieldIndex("qty"), self._float_delegate)
        self._float2_delegate = FloatDelegate(2, allow_tail=False)
        self._view.setItemDelegateForColumn(self.fieldIndex("fee"), self._float2_delegate)
        self._float4_delegate = FloatDelegate(4, allow_tail=False)
        self._view.setItemDelegateForColumn(self.fieldIndex("open_price"), self._float4_delegate)
        self._view.setItemDelegateForColumn(self.fieldIndex("close_price"), self._float4_delegate)
        self._profit_delegate = FloatDelegate(2, allow_tail=False, colors=True)
        self._view.setItemDelegateForColumn(self.fieldIndex("profit"), self._profit_delegate)
        self._view.setItemDelegateForColumn(self.fieldIndex("rel_profit"), self._profit_delegate)

    def setDatesRange(self, begin, end):
        self._begin = begin
        self._end = end
        self.calculateDealsReport()
        self.configureView()

    def setAccount(self, account_id):
        self._account_id = account_id
        self.calculateDealsReport()
        self.configureView()

    def setGrouping(self, group_dates):
        self._group_dates = group_dates
        self.calculateDealsReport()
        self.configureView()

    def calculateDealsReport(self):
        if self._account_id == 0:
            return
        if self._group_dates == 1:
            self._query = executeSQL(
                "SELECT asset, "
                "strftime('%s', datetime(open_timestamp, 'unixepoch', 'start of day')) as o_datetime, "
                "strftime('%s', datetime(close_timestamp, 'unixepoch', 'start of day')) as c_datetime, "
                "SUM(open_price*qty)/SUM(qty) as open_price, SUM(close_price*qty)/SUM(qty) AS close_price, "
                "SUM(qty) as qty, SUM(fee) as fee, SUM(profit) as profit, "
                "coalesce(100*SUM(qty*(close_price-open_price)-fee)/SUM(qty*open_price), 0) AS rel_profit "
                "FROM deals_ext "
                "WHERE account_id=:account_id AND close_timestamp>=:begin AND close_timestamp<=:end "
                "GROUP BY asset, o_datetime, c_datetime "
                "ORDER BY c_datetime, o_datetime",
                [(":account_id", self._account_id), (":begin", self._begin), (":end", self._end)], forward_only=False)
        else:
            self._query = executeSQL(
                "SELECT asset, open_timestamp AS o_datetime, close_timestamp AS c_datetime, "
                "open_price, close_price, qty, fee, profit, rel_profit, corp_action "
                "FROM deals_ext "
                "WHERE account_id=:account_id AND close_timestamp>=:begin AND close_timestamp<=:end "
                "ORDER BY c_datetime, o_datetime",
                [(":account_id", self._account_id), (":begin", self._begin), (":end", self._end)], forward_only=False)
        self.setQuery(self._query)
        self.modelReset.emit()


# ----------------------------------------------------------------------------------------------------------------------
class DealsReport(QObject):
    def __init__(self):
        super().__init__()
        self.name = self.name = self.tr("Deals by Account")
        self.window_class = "DealsReportWindow"


# ----------------------------------------------------------------------------------------------------------------------
class DealsReportWindow(MdiWidget, Ui_DealsReportWidget):
    def __init__(self, parent=None):
        MdiWidget.__init__(self, parent)
        self.setupUi(self)
        self.parent_mdi = parent

        self.category_model = DealsReportModel(self.ReportTableView)
        self.ReportTableView.setModel(self.category_model)

        self.connect_signals_and_slots()

    def connect_signals_and_slots(self):
        self.ReportRange.changed.connect(self.ReportTableView.model().setDatesRange)
        self.ReportAccountBtn.changed.connect(self.onAccountChange)
        self.ReportGroupCheck.clicked.connect(self.onGroupChange)

    @Slot()
    def onAccountChange(self):
        self.ReportTableView.model().setAccount(self.ReportAccountBtn.account_id)

    @Slot()
    def onGroupChange(self):
        group_dates = 1 if self.ReportGroupCheck.isChecked() else 0
        self.ReportTableView.model().setGrouping(group_dates)



import logging
from datetime import datetime
from PySide6.QtWidgets import QApplication

from jal.constants import Setup, BookAccount, PredefindedAccountType, AssetData, MarketDataFeed, PredefinedAsset
from jal.db.helpers import db_connection, executeSQL, readSQL, get_country_by_code


# ----------------------------------------------------------------------------------------------------------------------
class JalDB:
    def __init__(self):
        pass

    def tr(self, text):
        return QApplication.translate("JalDB", text)

    def commit(self):
        db_connection().commit()

    def get_language_id(self, language_code):
        return readSQL("SELECT id FROM languages WHERE language = :language_code", [(':language_code', language_code)])

    def get_language_code(self, language_id):
        return readSQL("SELECT language FROM languages WHERE id = :language_id", [(':language_id', language_id)])

    def get_asset_name(self, asset_id, full=False):
        if full:
            return readSQL("SELECT full_name FROM assets WHERE id=:asset_id", [(":asset_id", asset_id)])
        else:   # FIXME Below query may return several symbols
            return readSQL("SELECT symbol FROM assets AS a LEFT JOIN asset_tickers AS s "   
                           "ON s.asset_id=a.id AND s.active=1 WHERE a.id=:asset_id", [(":asset_id", asset_id)])

    def get_asset_type(self, asset_id):
        return readSQL("SELECT type_id FROM assets WHERE id=:asset_id", [(":asset_id", asset_id)])

    def get_asset_country(self, asset_id):
        return readSQL("SELECT c.name FROM assets AS a LEFT JOIN countries AS c ON c.id=a.country_id "
                       "WHERE a.id=:asset_id", [(":asset_id", asset_id)])

    def get_account_name(self, account_id):
        return readSQL("SELECT name FROM accounts WHERE id=:account_id", [(":account_id", account_id)])

    # Searches for account_id by account number and optional currency
    # Returns: account_id or None if no account was found
    def get_account_id(self, accountNumber, accountCurrency=''):
        if accountCurrency:
            account_id = readSQL("SELECT a.id FROM accounts AS a "
                                 "LEFT JOIN currencies AS c ON c.id=a.currency_id "
                                 "WHERE a.number=:account_number AND c.symbol=:currency_name",
                                 [(":account_number", accountNumber), (":currency_name", accountCurrency)],
                                 check_unique=True)
        else:
            account_id = readSQL("SELECT a.id FROM accounts AS a WHERE a.number=:account_number",
                                 [(":account_number", accountNumber)],
                                 check_unique=True)
        return account_id

    def find_account(self, account_number, currency_code):
        return readSQL("SELECT id FROM accounts WHERE number=:account_number AND currency_id=:currency",
                       [(":account_number", account_number), (":currency", currency_code)], check_unique=True)

    def add_account(self, account_number, currency_id, account_type=PredefindedAccountType.Investment):
        account_id = self.find_account(account_number, currency_id)
        if account_id:  # Account already exists
            logging.warning(self.tr("Account already exists: ") +
                            f"{account_number} ({self.get_asset_name(currency_id)})")
            return account_id
        currency = self.get_asset_name(currency_id)
        account = readSQL("SELECT a.name, a.organization_id, a.country_id, c.symbol AS currency FROM accounts a "
                          "LEFT JOIN assets_ext c ON c.id=a.currency_id WHERE a.number=:number LIMIT 1",
                          [(":number", account_number)], named=True)
        if account:  # Account with the same number but different currency exists
            if account['name'][-len(account['currency']):] == account['currency']:
                new_name = account['name'][:-len(account['currency'])] + currency
            else:
                new_name = account['name'] + '.' + currency
            query = executeSQL(
                "INSERT INTO accounts (type_id, name, active, number, currency_id, organization_id, country_id) "
                "SELECT a.type_id, :new_name, a.active, a.number, :currency_id, a.organization_id, a.country_id "
                "FROM accounts AS a LEFT JOIN assets AS c ON c.id=:currency_id "
                "WHERE number=:account_number LIMIT 1",
                [(":account_number", account_number), (":currency_id", currency_id), (":new_name", new_name)])
            return query.lastInsertId()

        bank_name = self.tr("Bank for #" + account_number)
        bank_id = readSQL("SELECT id FROM agents WHERE name=:bank_name", [(":bank_name", bank_name)])
        if bank_id is None:
            query = executeSQL("INSERT INTO agents (pid, name) VALUES (0, :bank_name)", [(":bank_name", bank_name)])
            bank_id = query.lastInsertId()
        query = executeSQL("INSERT INTO accounts (type_id, name, active, number, currency_id, organization_id) "
                           "VALUES(:type, :name, 1, :number, :currency, :bank)",
                           [(":type", account_type), (":name", account_number+'.'+currency),
                            (":number", account_number), (":currency", currency_id), (":bank", bank_id)])
        return query.lastInsertId()

    def get_account_currency(self, account_id):
        return readSQL("SELECT currency_id FROM accounts WHERE id=:account_id", [(":account_id", account_id)])

    def get_account_bank(self, account_id):
        return readSQL("SELECT organization_id FROM accounts WHERE id=:account_id", [(":account_id", account_id)])

    # Searches for asset_id in database based on keys available in search data:
    # first by 'isin', then by 'reg_number', next by 'symbol' and other
    # Returns: asset_id or None if not found
    def get_asset_id(self, search_data):
        asset_id = None
        if 'isin' in search_data and search_data['isin']:
            asset_id = readSQL("SELECT id FROM assets WHERE isin=:isin", [(":isin", search_data['isin'])])
        if asset_id is None and 'reg_number' in search_data and search_data['reg_number']:
            asset_id = readSQL("SELECT asset_id FROM asset_data WHERE datatype=:datatype AND value=:reg_number",
                               [(":datatype", AssetData.RegistrationCode), (":reg_number", search_data['reg_number'])])
        if asset_id is None and 'symbol' in search_data and search_data['symbol']:
            if 'type' in search_data:
                if 'expiry' in search_data:
                    asset_id = readSQL("SELECT a.id FROM assets_ext a "
                                       "LEFT JOIN asset_data d ON a.id=d.asset_id AND d.datatype=:datatype "
                                       "WHERE symbol=:symbol COLLATE NOCASE AND type_id=:type AND value=:value",
                                       [(":datatype", AssetData.ExpiryDate), (":symbol", search_data['symbol']),
                                        (":type", search_data['type']), (":expiry", search_data['expiry'])])
                else:
                    asset_id = readSQL("SELECT id FROM assets_ext "
                                       "WHERE symbol=:symbol COLLATE NOCASE and type_id=:type",
                                       [(":symbol", search_data['symbol']), (":type", search_data['type'])])
            else:
                asset_id = readSQL("SELECT id FROM assets_ext WHERE symbol=:symbol COLLATE NOCASE",
                                   [(":symbol", search_data['symbol'])])
        return asset_id

    def get_quote(self, asset_id, currency_id, timestamp):
        return readSQL("SELECT quote FROM quotes WHERE asset_id=:asset_id "
                       "AND currency_id=:currency_id AND timestamp=:timestamp",
                       [(":asset_id", asset_id), (":currency_id", currency_id), (":timestamp", timestamp)])

    # Set quotations for given asset_id and currency_id
    # quotations is a list of {'timestamp', 'quote'} values
    def update_quotes(self, asset_id, currency_id, quotations):
        data = [x for x in quotations if x['timestamp'] is not None and x['quote'] is not None]  # Drop Nones
        if data:
            for quote in quotations:
                _ = executeSQL("INSERT OR REPLACE INTO quotes (asset_id, currency_id, timestamp, quote) "
                               "VALUES(:asset_id, :currency_id, :timestamp, :quote)",
                               [(":asset_id", asset_id), (":currency_id", currency_id),
                                (":timestamp", quote['timestamp']), (":quote", quote['quote'])])
            begin = min(data, key=lambda x: x['timestamp'])['timestamp']
            end = max(data, key=lambda x: x['timestamp'])['timestamp']
            logging.info(self.tr("Quotations were updated: ") +
                         f"{self.get_asset_name(asset_id)} ({self.get_asset_name(asset_id)}) " 
                         f"{datetime.utcfromtimestamp(begin).strftime('%d/%m/%Y')} - "
                         f"{datetime.utcfromtimestamp(end).strftime('%d/%m/%Y')}")

    def add_asset(self, asset_type, name, isin, country_code=''):
        country_id = get_country_by_code(country_code)
        query = executeSQL("INSERT INTO assets (type_id, full_name, isin, country_id) "
                           "VALUES (:type, :full_name, :isin, :country_id)",
                           [(":type", asset_type), (":full_name", name),
                            (":isin", isin), (":country_id", country_id)], commit=True)
        return query.lastInsertId()

    def add_symbol(self, asset_id, symbol, currency, note, data_source=MarketDataFeed.NA):
        existing = readSQL("SELECT id, symbol, description, quote_source FROM asset_tickers "
                           "WHERE asset_id=:asset_id AND symbol=:symbol AND currency_id=:currency",
                           [(":asset_id", asset_id), (":symbol", symbol), (":currency", currency)], named=True)
        if existing is None:   # Create new symbol
            _ = executeSQL("INSERT INTO asset_tickers (asset_id, symbol, currency_id, description, quote_source) "
                           "VALUES (:asset_id, :symbol, :currency, :note, :data_source)",
                           [(":asset_id", asset_id), (":symbol", symbol), (":currency", currency),
                            (":note", note), (":data_source", data_source)])
        else:   # Update data for existing symbol
            if not existing['description']:
                _ = executeSQL("UPDATE asset_tickers SET description=:note WHERE id=:id",
                               [(":note", note), (":id", existing['id'])])
            if existing['quote_source'] == MarketDataFeed.NA:
                _ = executeSQL("UPDATE asset_tickers SET quote_source=:data_source WHERE id=:id",
                               [(":data_source", data_source), (":id", existing['id'])])

    def update_asset_data(self, asset_id, data):
        asset = readSQL("SELECT type_id, isin, full_name FROM assets WHERE id=:asset_id",
                        [(":asset_id", asset_id)], named=True)
        if asset is None:
            logging.warning(self.tr("Asset not found for update: ") + f"{data}")
            return
        if 'isin' in data and data['isin']:
            if asset['isin']:
                if asset['isin'] != data['isin']:
                    logging.error(self.tr("Unexpected attempt to update ISIN for ")
                                  + f"{self.get_asset_name(asset_id)}: {asset['isin']} -> {data['isin']}")
            else:
                _ = executeSQL("UPDATE assets SET isin=:new_isin WHERE id=:asset_id",
                               [(":new_isin", data['isin']), (":asset_id", asset_id)])
        if 'name' in data and data['name']:
            if not asset['full_name']:
                _ = executeSQL("UPDATE assets SET full_name=:new_name WHERE id=:asset_id",
                               [(":new_name", data['name']), (":asset_id", asset_id)])
        if 'country' in data and data['country']:
            country_id, country_code = readSQL("SELECT a.country_id, c.code FROM assets AS a LEFT JOIN countries AS c "
                                               "ON a.country_id=c.id WHERE a.id=:asset_id", [(":asset_id", asset_id)])
            if (country_id == 0) or (country_code.lower() != data['country'].lower()):
                new_country_id = get_country_by_code(data['country'])
                _ = executeSQL("UPDATE assets SET country_id=:new_country_id WHERE id=:asset_id",
                               [(":new_country_id", new_country_id), (":asset_id", asset_id)])
                if country_id != 0:
                    logging.info(self.tr("Country updated for ")
                                 + f"{self.get_asset_name(asset_id)}: {country_code} -> {data['country']}")
        if 'reg_number' in data and data['reg_number']:
            reg_number = readSQL("SELECT value FROM asset_data WHERE datatype=:datatype AND asset_id=:asset_id",
                                 [(":datatype", AssetData.RegistrationCode), (":asset_id", asset_id)])
            if reg_number != data['reg_number']:
                _ = executeSQL("INSERT OR REPLACE INTO asset_data(asset_id, datatype, value) "
                               "VALUES(:asset_id, :datatype, :reg_number)",
                               [(":asset_id", asset_id), (":datatype", AssetData.RegistrationCode),
                                (":reg_number", data['reg_number'])])
                if reg_number:
                    logging.info(self.tr("Reg.number updated for ")
                                 + f"{self.get_asset_name(asset_id)}: {reg_number} -> {data['reg_number']}")
        if 'expiry' in data and data['expiry']:
            _ = executeSQL("INSERT OR REPLACE INTO asset_data(asset_id, datatype, value) "
                           "VALUES(:asset_id, :datatype, :expiry)",
                           [(":asset_id", asset_id), (":datatype", AssetData.ExpiryDate),
                            (":expiry", data['expiry'])])
        if 'principal' in data and asset['type_id'] == PredefinedAsset.Bond:
            try:
                principal = float(data['principal'])
                if principal > 0:
                    _ = executeSQL("INSERT OR REPLACE INTO asset_data(asset_id, datatype, value) "
                                   "VALUES(:asset_id, :datatype, :principal)",
                                   [(":asset_id", asset_id), (":datatype", AssetData.PrincipalValue),
                                    (":principal", principal)])
            except ValueError:
                pass

    def add_dividend(self, subtype, timestamp, account_id, asset_id, amount, note, trade_number='',
                     tax=0.0, price=None):
        id = readSQL("SELECT id FROM dividends WHERE timestamp=:timestamp AND type=:subtype AND account_id=:account_id "
                     "AND asset_id=:asset_id AND amount=:amount AND note=:note",
                     [(":timestamp", timestamp), (":subtype", subtype), (":account_id", account_id),
                      (":asset_id", asset_id), (":amount", amount), (":note", note)])
        if id:
            logging.info(self.tr("Dividend already exists: ") + f"{note}")
            return
        _ = executeSQL("INSERT INTO dividends (timestamp, number, type, account_id, asset_id, amount, tax, note) "
                       "VALUES (:timestamp, :number, :subtype, :account_id, :asset_id, :amount, :tax, :note)",
                       [(":timestamp", timestamp), (":number", trade_number), (":subtype", subtype),
                        (":account_id", account_id), (":asset_id", asset_id), (":amount", amount),
                        (":tax", tax), (":note", note)],
                       commit=True)
        if price is not None:
            self.update_quotes(asset_id, self.get_account_currency(account_id),
                               [{'timestamp': timestamp, 'quote': price}])

    def update_dividend_tax(self, dividend_id, new_tax):
        _ = executeSQL("UPDATE dividends SET tax=:tax WHERE id=:dividend_id",
                       [(":dividend_id", dividend_id), (":tax", new_tax)], commit=True)

    def add_trade(self, account_id, asset_id, timestamp, settlement, number, qty, price, fee, note=''):
        trade_id = readSQL("SELECT id FROM trades "
                           "WHERE timestamp=:timestamp AND asset_id = :asset "
                           "AND account_id = :account AND number = :number AND qty = :qty AND price = :price",
                           [(":timestamp", timestamp), (":asset", asset_id), (":account", account_id),
                            (":number", number), (":qty", qty), (":price", price)])
        if trade_id:
            logging.info(self.tr("Trade already exists: #") + f"{number}")
            return

        _ = executeSQL("INSERT INTO trades (timestamp, settlement, number, account_id, asset_id, qty, price, fee, note)"
                       " VALUES (:timestamp, :settlement, :number, :account, :asset, :qty, :price, :fee, :note)",
                       [(":timestamp", timestamp), (":settlement", settlement), (":number", number),
                        (":account", account_id), (":asset", asset_id), (":qty", float(qty)),
                        (":price", float(price)), (":fee", float(fee)), (":note", note)], commit=True)

    def del_trade(self, account_id, asset_id, timestamp, _settlement, number, qty, price, _fee):
        _ = executeSQL("DELETE FROM trades "
                       "WHERE timestamp=:timestamp AND asset_id=:asset "
                       "AND account_id=:account AND number=:number AND qty=:qty AND price=:price",
                       [(":timestamp", timestamp), (":asset", asset_id), (":account", account_id),
                        (":number", number), (":qty", -qty), (":price", price)], commit=True)

    def add_transfer(self, timestamp, f_acc_id, f_amount, t_acc_id, t_amount, fee_acc_id, fee, note):
        transfer_id = readSQL("SELECT id FROM transfers WHERE withdrawal_timestamp=:timestamp "
                              "AND withdrawal_account=:from_acc_id AND deposit_account=:to_acc_id "
                              "AND withdrawal=:f_amount AND deposit=:t_amount",
                              [(":timestamp", timestamp), (":from_acc_id", f_acc_id), (":to_acc_id", t_acc_id),
                               (":f_amount", f_amount), (":t_amount", t_amount)])
        if transfer_id:
            logging.info(self.tr("Transfer/Exchange already exists: ") + f"{f_amount}->{t_amount}")
            return
        if abs(fee) > Setup.CALC_TOLERANCE:
            _ = executeSQL("INSERT INTO transfers (withdrawal_timestamp, withdrawal_account, withdrawal, "
                           "deposit_timestamp, deposit_account, deposit, fee_account, fee, note) "
                           "VALUES (:timestamp, :f_acc_id, :f_amount, :timestamp, :t_acc_id, :t_amount, "
                           ":fee_acc_id, :fee_amount, :note)",
                           [(":timestamp", timestamp), (":f_acc_id", f_acc_id), (":t_acc_id", t_acc_id),
                            (":f_amount", f_amount), (":t_amount", t_amount), (":fee_acc_id", fee_acc_id),
                            (":fee_amount", fee), (":note", note)], commit=True)
        else:
            _ = executeSQL("INSERT INTO transfers (withdrawal_timestamp, withdrawal_account, withdrawal, "
                           "deposit_timestamp, deposit_account, deposit, note) "
                           "VALUES (:timestamp, :f_acc_id, :f_amount, :timestamp, :t_acc_id, :t_amount, :note)",
                           [(":timestamp", timestamp), (":f_acc_id", f_acc_id), (":t_acc_id", t_acc_id),
                            (":f_amount", f_amount), (":t_amount", t_amount), (":note", note)], commit=True)

    def add_corporate_action(self, account_id, type, timestamp, number,
                             asset_id_old, qty_old, asset_id_new, qty_new, basis_ratio, note):
        action_id = readSQL("SELECT id FROM corp_actions "
                            "WHERE timestamp=:timestamp AND type = :type AND account_id = :account AND number = :number "
                            "AND asset_id = :asset AND asset_id_new = :asset_new",
                            [(":timestamp", timestamp), (":type", type), (":account", account_id), (":number", number),
                             (":asset", asset_id_old), (":asset_new", asset_id_new)])
        if action_id:
            logging.info(self.tr("Corporate action already exists: #") + f"{number}")
            return

        _ = executeSQL("INSERT INTO corp_actions (timestamp, number, account_id, type, "
                       "asset_id, qty, asset_id_new, qty_new, basis_ratio, note) "
                       "VALUES (:timestamp, :number, :account, :type, "
                       ":asset, :qty, :asset_new, :qty_new, :basis_ratio, :note)",
                       [(":timestamp", timestamp), (":number", number), (":account", account_id), (":type", type),
                        (":asset", asset_id_old), (":qty", float(qty_old)), (":asset_new", asset_id_new),
                        (":qty_new", float(qty_new)), (":basis_ratio", basis_ratio), (":note", note)], commit=True)

    def add_cash_transaction(self, account_id, broker_id, timestamp, amount, category_id, description):
        query = executeSQL("INSERT INTO actions (timestamp, account_id, peer_id) "
                           "VALUES (:timestamp, :account_id, :bank_id)",
                           [(":timestamp", timestamp), (":account_id", account_id), (":bank_id", broker_id)])
        pid = query.lastInsertId()
        _ = executeSQL("INSERT INTO action_details (pid, category_id, amount, note) "
                       "VALUES (:pid, :category_id, :amount, :note)",
                       [(":pid", pid), (":category_id", category_id), (":amount", amount),
                        (":note", description)], commit=True)

    def reconcile_account(self, account_id, timestamp):
        _ = executeSQL("UPDATE accounts SET reconciled_on=:timestamp WHERE id = :account_id",
                       [(":timestamp", timestamp), (":account_id", account_id)])

    def account_reconciliation_timestamp(self, account_id):
        timestamp = readSQL("SELECT reconciled_on FROM accounts WHERE id=:account_id", [(":account_id", account_id)])
        if timestamp is None:
            return 0
        else:
            return timestamp

    def get_asset_amount(self, timestamp, account_id, asset_id):
        return readSQL("SELECT amount_acc FROM ledger "
                       "WHERE account_id=:account_id AND asset_id=:asset_id AND timestamp<=:timestamp "
                       "AND (book_account=:money OR book_account=:assets)"
                       "ORDER BY id DESC LIMIT 1",
                       [(":account_id", account_id), (":asset_id", asset_id), (":timestamp", timestamp),
                        (":money", BookAccount.Money), (":assets", BookAccount.Assets)])



import logging
from datetime import datetime, timezone
from lxml import etree
from PySide6.QtWidgets import QApplication
from jal.data_import.statement import Statement, FOF, Statement_ImportError


# -----------------------------------------------------------------------------------------------------------------------
# Base class to load XML-based statements
class StatementXML(Statement):
    statements_path = ''    # Where in XML structure search for statements
    statement_tag = ''      # Tag of the statement in XML (there might be several statements in one XML)
    STATEMENT_ROOT = '<statement_root>'

    def __init__(self):
        super().__init__()
        self.statement_name = ''
        self._sections = {}
        self._data = {
            FOF.PERIOD: [None, None],
            FOF.ACCOUNTS: [],
            FOF.ASSETS: [],
            FOF.SYMBOLS: [],
            FOF.ASSETS_DATA: [],
            FOF.TRADES: [],
            FOF.TRANSFERS: [],
            FOF.CORP_ACTIONS: [],
            FOF.ASSET_PAYMENTS: [],
            FOF.INCOME_SPENDING: []
        }
        self.attr_loader = {
            str: self.attr_string,
            float: self.attr_number,
            datetime: self.attr_timestamp
        }

    # -----------------------------------------------------------------------------------------------------------------------
    # Helpers to get values from XML tag properties
    # Convert attribute 'attr_name' value to string or return default value if attribute not found
    @staticmethod
    def attr_string(xml_element, attr_name, default_value):
        if attr_name not in xml_element.attrib:
            return default_value
        return xml_element.attrib[attr_name].strip()

    # Convert attribute 'attr_name' value to float or return default value if attribute not found / not a number
    @staticmethod
    def attr_number(xml_element, attr_name, default_value):
        if attr_name not in xml_element.attrib:
            return default_value
        try:
            value = float(xml_element.attrib[attr_name])
        except ValueError:
            return None
        return value

    # Convert attribute 'attr_name' value from strings "YYYYMMDD:hhmmss' or "YYYYMMDD" to datetime object
    # or return default value if attribute not found / has wrong format
    @staticmethod
    def attr_timestamp(xml_element, attr_name, default_value):
        if attr_name not in xml_element.attrib:
            return default_value
        time_str = xml_element.attrib[attr_name]
        try:
            if len(time_str) == 19:  # YYYY-MM-DDTHH:MM:SS
                return int(datetime.strptime(time_str, "%Y-%m-%dT%H:%M:%S").replace(tzinfo=timezone.utc).timestamp())
            if len(time_str) == 15:  # YYYYMMDD;HHMMSS
                return int(datetime.strptime(time_str, "%Y%m%d;%H%M%S").replace(tzinfo=timezone.utc).timestamp())
            elif len(time_str) == 8:  # YYYYMMDD
                return int(datetime.strptime(time_str, "%Y%m%d").replace(tzinfo=timezone.utc).timestamp())
            else:
                return default_value
        except ValueError:
            raise Statement_ImportError(QApplication.translate("StatementXML", "Unsupported date/time format: ")
                                        + f"{xml_element.attrib[attr_name]}")

    def load(self, filename: str) -> None:
        try:
            xml_root = etree.parse(filename)
        except etree.XMLSyntaxError as e:
            raise Statement_ImportError(self.tr("Can't parse XML file: ") + e.msg)
        self.validate_file_header_attributes(xml_root.findall('.')[0].attrib)
        statements = xml_root.findall(self.statements_path)
        for statement in statements:
            if statement.tag != self.statement_tag:
                continue
            header_data = self.get_section_data(statement)
            self._sections[StatementXML.STATEMENT_ROOT]['loader'](header_data)

            for section in self._sections:
                if section == StatementXML.STATEMENT_ROOT:
                    continue  # skip header description
                section_elements = statement.xpath(section)  # Actually should be list of 0 or 1 element
                if section_elements:
                    section_data = self.get_section_data(section_elements[0])
                    if section_data is None:
                        return
                    self._sections[section]['loader'](section_data)
        logging.info(self.statement_name + self.tr(" loaded successfully"))

    def validate_file_header_attributes(self, xml_data):
        return

    def get_section_data(self, section):
        if section.tag == self.statement_tag:  # This is header section
            return self.parse_attributes(StatementXML.STATEMENT_ROOT, section)
        try:
            tag = self._sections[section.tag]['tag']
        except KeyError:
            return []  # This section isn't used for import
        data = []
        for element in section.xpath(tag):
            attributes = self.parse_attributes(section.tag, element)
            if attributes is not None:
                data.append(attributes)
        return data

    def parse_attributes(self, section_tag, element):
        tag_dictionary = {}
        if self._sections[section_tag]['level']:  # Skip extra lines (SUMMARY, etc)
            if self.attr_string(element, 'levelOfDetail', '') != self._sections[section_tag]['level']:
                return None
        for attr_name, key_name, attr_type, attr_default in self._sections[section_tag]['values']:
            attr_value = self.attr_loader[attr_type](element, attr_name, attr_default)
            if attr_value is None:
                raise Statement_ImportError(self.tr("Failed to load attribute: ") + f"{attr_name} / {element.attrib}")
            tag_dictionary[key_name] = attr_value
        return tag_dictionary

    # Removes all keys listed in extra_keys_list from operation_dict
    def drop_extra_fields(self, operation_dict, extra_keys_list):
        for key in extra_keys_list:
            if key in operation_dict:
                del operation_dict[key]



import logging
from datetime import datetime, timezone

from jal.constants import Setup
from jal.data_import.statement import FOF
from jal.data_import.statement_xls import StatementXLS

JAL_STATEMENT_CLASS = "StatementPSB"


# ----------------------------------------------------------------------------------------------------------------------
class StatementPSB(StatementXLS):
    Header = (2, 3, 'Брокер: ПАО "Промсвязьбанк"')
    PeriodPattern = (3, 6, r"с (?P<S>\d\d\.\d\d\.\d\d\d\d) по (?P<E>\d\d\.\d\d\.\d\d\d\d)")
    AccountPattern = (3, 9, r"(?P<ACCOUNT>\S*)( от \d\d\.\d\d\.\d\d\d\d)?")
    HeaderCol = 1
    SummaryHeader = "Сводная информация по счетам клиента в валюте счета"
    money_section = "Позиция денежных средств по биржевым площадкам"
    money_columns = {
        "name": "Валюта",
        "begin": "Входящие лимиты",
        "end": "Плановый исходящий остаток",
        "settled_end": "Исходящие лимиты"
    }
    asset_section = r"^Портфель на конец дня.*"
    asset_columns = {
        "name": r"Наименование эмитента, вид, категория \(тип\), выпуск, транш ЦБ",
        "isin": "ISIN",
        "reg_number": r"Номер гос\.регистрации"
    }

    def __init__(self):
        super().__init__()
        self.name = self.tr("PSB Broker")
        self.icon_name = "psb.ico"
        self.filename_filter = self.tr("PSB broker statement (*.xlsx *.xls)")

    def _load_deals(self):
        cnt = 0
        sections = [
            r"Сделки, .* с ЦБ на биржевых торговых .* с расчетами в дату заключения",
            r"Сделки, .* с ЦБ на биржевых торговых .* с расчетами Т\+, незавершенные в отчетном периоде",
            r"Сделки, .* с ЦБ на биржевых торговых .* с расчетами Т\+, рассчитанные в отчетном периоде"
        ]
        columns = {
            "timestamp": "Дата и время совершения сделки",
            "*settlement": r".*ая дата исполнения сделки",
            "number": "Номер сделки в ТС",
            "isin": "ISIN",
            "reg_number": r"Номер гос\. регистрации",
            "B/S": r"Вид сделки \(покупка\/продажа\)",
            "qty": r"Кол-во ЦБ, шт\.",
            "currency": r"Валюта сделки \/ Валюта платежа",
            "price": r"Цена \(% для обл\)",
            "amount": "Сумма сделки без НКД",
            "accrued_int": "НКД",
            "fee1": r"Комиссия торговой системы( \(без НДС\))?, руб",
            "fee2": r"Клиринговая комиссия( \(без НДС\))?, руб",
            "fee3": r"Комиссия за ИТС( \(в т.ч. НДС\))?, руб",
            "fee_broker": r"Ком\. брокера"
        }
        for section in sections:
            row, headers = self.find_section_start(section, columns)
            if row < 0:
                continue
            while row < self._statement.shape[0]:
                if self._statement[self.HeaderCol][row].startswith('Итого') or \
                        self._statement[self.HeaderCol][row] == '':
                    break
                deal_number = self._statement[headers['number']][row]
                deal_currency = self._statement[headers['currency']][row].split('/')[0].strip()    # <- 'deal / payment'
                try:
                    code = self.currency_id(self.currency_substitutions[deal_currency])
                except KeyError:
                    code = self.currency_id(deal_currency)
                asset_id = self.asset_id({'isin': self._statement[headers['isin']][row],
                                          'reg_number': self._statement[headers['reg_number']][row],
                                          'currency': code, 'search_online': "MOEX"})
                if self._statement[headers['B/S']][row] == 'покупка':
                    qty = self._statement[headers['qty']][row]
                    bond_interest = -self._statement[headers['accrued_int']][row]
                elif self._statement[headers['B/S']][row] == 'продажа':
                    qty = -self._statement[headers['qty']][row]
                    bond_interest = self._statement[headers['accrued_int']][row]
                else:
                    row += 1
                    logging.warning(self.tr("Unknown trade type: ") + self._statement[headers['B/S']][row])
                    continue

                price = self._statement[headers['price']][row]
                currencies = [x.strip() for x in self._statement[headers['currency']][row].split('/')]
                if currencies[0] != currencies[1]:
                    row += 1
                    logging.warning(self.tr("Unsupported trade with different currencies: ") + currencies)
                    continue
                currency = currencies[0]
                fee = self._statement[headers['fee1']][row] + self._statement[headers['fee2']][row] + \
                      self._statement[headers['fee3']][row] + self._statement[headers['fee_broker']][row]
                amount = self._statement[headers['amount']][row]
                if abs(abs(price * qty) - amount) >= Setup.DISP_TOLERANCE:
                    price = abs(amount / qty)
                timestamp = int(datetime.strptime(self._statement[headers['timestamp']][row],
                                                  "%d.%m.%Y %H:%M:%S").replace(tzinfo=timezone.utc).timestamp())
                if headers['*settlement'] == -1:
                    settlement = int(
                        datetime.strptime(self._statement[headers['timestamp']][row], "%d.%m.%Y %H:%M:%S").replace(
                            tzinfo=timezone.utc).replace(hour=0, minute=0, second=0).timestamp())
                else:
                    settlement = int(datetime.strptime(self._statement[headers['*settlement']][row],
                                                       "%d.%m.%Y").replace(tzinfo=timezone.utc).timestamp())
                account_id = self._find_account_id(self._account_number, currency)
                new_id = max([0] + [x['id'] for x in self._data[FOF.TRADES]]) + 1
                trade = {"id": new_id, "number": deal_number, "timestamp": timestamp, "settlement": settlement,
                         "account": account_id, "asset": asset_id, "quantity": qty, "price": price, "fee": fee}
                self._data[FOF.TRADES].append(trade)
                if bond_interest != 0:
                    new_id = max([0] + [x['id'] for x in self._data[FOF.ASSET_PAYMENTS]]) + 1
                    payment = {"id": new_id, "type": FOF.PAYMENT_INTEREST, "account": account_id,
                               "timestamp": timestamp,
                               "number": deal_number, "asset": asset_id, "amount": bond_interest, "description": "НКД"}
                    self._data[FOF.ASSET_PAYMENTS].append(payment)
                cnt += 1
                row += 1
        logging.info(self.tr("Trades loaded: ") + f"{cnt}")

    def _load_cash_transactions(self):
        self.load_cash_transactions()
        self.load_coupons()
        self.load_dividends()

    def load_cash_transactions(self):
        cnt = 0
        columns = {
            "date": "Дата",
            "currency": "Валюта счета",
            "amount": "Сумма",
            "operation": "Операция",
            "note": "Комментарий"
        }

        row, headers = self.find_section_start("Внешнее движение денежных средств в валюте счета", columns)
        if row < 0:
            return False
        while row < self._statement.shape[0]:
            if self._statement[self.HeaderCol][row] == '':
                break
            if self._statement[headers['note']][row].startswith("Дивиденды") or \
                    self._statement[headers['note']][row].startswith("Погашение купона"):
                row += 1   # These data present in separate tables to load
                continue
            if self._statement[headers['note']][row] != '':
                logging.warning(self.tr("Unknown cash transaction: ") + self._statement[headers['note']][row])
                row += 1
                continue

            timestamp = int(datetime.strptime(self._statement[headers['date']][row],
                                              "%d.%m.%Y").replace(tzinfo=timezone.utc).timestamp())
            amount = self._statement[headers['amount']][row]
            account_id = self._find_account_id(self._account_number, self._statement[headers['currency']][row])
            if self._statement[headers['operation']][row] == 'Зачислено на счет':
                self.transfer_in(timestamp, account_id, amount)
            elif self._statement[headers['operation']][row] == 'Списано со счета':
                self.transfer_out(timestamp, account_id, amount)
            else:
                logging.warning(self.tr("Unknown cash operation: ") + self._statement[headers['operation']][row])
                row += 1
                continue
            cnt += 1
            row += 1
        logging.info(self.tr("Cash transactions loaded: ") + f"{cnt}")

    def transfer_in(self, timestamp, account_id, amount):
        account = [x for x in self._data[FOF.ACCOUNTS] if x["id"] == account_id][0]
        new_id = max([0] + [x['id'] for x in self._data[FOF.TRANSFERS]]) + 1
        transfer = {"id": new_id, "account": [0, account_id, 0],
                    "asset": [account['currency'], account['currency']], "timestamp": timestamp,
                    "withdrawal": amount, "deposit": amount, "fee": 0.0}
        self._data[FOF.TRANSFERS].append(transfer)

    def transfer_out(self, timestamp, account_id, amount):
        account = [x for x in self._data[FOF.ACCOUNTS] if x["id"] == account_id][0]
        new_id = max([0] + [x['id'] for x in self._data[FOF.TRANSFERS]]) + 1
        transfer = {"id": new_id, "account": [account_id, 0, 0],
                    "asset": [account['currency'], account['currency']], "timestamp": timestamp,
                    "withdrawal": -amount, "deposit": -amount, "fee": 0.0}
        self._data[FOF.TRANSFERS].append(transfer)

    def load_coupons(self):
        cnt = 0
        columns = {
            "date": "Дата операции",
            "operation": "Вид операции",
            "asset_name": r"Наименование эмитента, вид, категория \(тип\), выпуск, транш ЦБ",
            "isin": "ISIN",
            "reg_number": "Регистрационный номер ЦБ",
            "currency": "Валюта Выплаты",
            "coupon": "НКД",
            "tax": r"Сумма удержанного налога, руб.*"
        }

        row, headers = self.find_section_start("Погашение купонов и ЦБ", columns)
        if row < 0:
            return False
        while row < self._statement.shape[0]:
            if self._statement[2][row] == '':  # Row may contain comment in cell [1]
                break
            if self._statement[headers['operation']][row] != 'Погашение купона':
                logging.warning(self.tr("Unsupported payment: ") + self._statement[headers['operation']][row])
                row += 1
                continue
            timestamp = int(datetime.strptime(self._statement[headers['date']][row],
                                              "%d.%m.%Y").replace(tzinfo=timezone.utc).timestamp())
            amount = float(self._statement[headers['coupon']][row])
            tax = float(self._statement[headers['tax']][row])
            account_id = self._find_account_id(self._account_number, self._statement[headers['currency']][row])
            try:
                code = self.currency_id(self.currency_substitutions[self._statement[headers['currency']][row]])
            except KeyError:
                code = self.currency_id(self._statement[headers['currency']][row])
            asset_id = self.asset_id({'isin': self._statement[headers['isin']][row],
                                      'reg_number': self._statement[headers['reg_number']][row],
                                      'currency': code, 'search_online': "MOEX"})
            note = self._statement[headers['operation']][row] + " " + self._statement[headers['asset_name']][row]
            new_id = max([0] + [x['id'] for x in self._data[FOF.ASSET_PAYMENTS]]) + 1
            payment = {"id": new_id, "type": FOF.PAYMENT_INTEREST, "account": account_id, "timestamp": timestamp,
                       "asset": asset_id, "amount": amount, "tax": tax, "description": note}
            self._data[FOF.ASSET_PAYMENTS].append(payment)
            cnt += 1
            row += 1
        logging.info(self.tr("Bond interests loaded: ") + f"{cnt}")

    def load_dividends(self):
        cnt = 0
        columns = {
            "date": "Дата операции",
            "isin": "ISIN",
            "reg_number": r"Номер гос\. регистрации",
            "currency": "Валюта Выплаты",
            "amount": "Сумма дивидендов",
            "tax": "Сумма удержанного налога"
        }

        row, headers = self.find_section_start("Выплата дивидендов", columns)
        if row < 0:
            return False
        while row < self._statement.shape[0]:
            if self._statement[self.HeaderCol][row] == '':
                break

            timestamp = int(datetime.strptime(self._statement[headers['date']][row],
                                              "%d.%m.%Y").replace(tzinfo=timezone.utc).timestamp())
            amount = float(self._statement[headers['amount']][row])
            tax = float(self._statement[headers['tax']][row])
            account_id = self._find_account_id(self._account_number, self._statement[headers['currency']][row])
            try:
                code = self.currency_id(self.currency_substitutions[self._statement[headers['currency']][row]])
            except KeyError:
                code = self.currency_id(self._statement[headers['currency']][row])
            asset_id = self.asset_id({'isin': self._statement[headers['isin']][row],
                                      'reg_number': self._statement[headers['reg_number']][row],
                                      'currency': code, 'search_online': "MOEX"})
            new_id = max([0] + [x['id'] for x in self._data[FOF.ASSET_PAYMENTS]]) + 1
            payment = {"id": new_id, "type": FOF.PAYMENT_DIVIDEND, "account": account_id, "timestamp": timestamp,
                       "asset": asset_id, "amount": amount, "tax": tax, "description": ''}
            self._data[FOF.ASSET_PAYMENTS].append(payment)
            cnt += 1
            row += 1
        logging.info(self.tr("Dividends loaded: ") + f"{cnt}")



from PySide6.QtCore import Qt, Slot, QObject
from PySide6.QtSql import QSqlTableModel
from jal.ui.reports.ui_profit_loss_report import Ui_ProfitLossReportWidget
from jal.db.helpers import db_connection, executeSQL
from jal.constants import BookAccount, PredefinedCategory
from jal.widgets.delegates import FloatDelegate, TimestampDelegate
from jal.widgets.mdi import MdiWidget

JAL_REPORT_CLASS = "ProfitLossReport"


#-----------------------------------------------------------------------------------------------------------------------
class ProfitLossReportModel(QSqlTableModel):
    def __init__(self, parent_view):
        self._columns = [("period", self.tr("Period")),
                         ("transfer", self.tr("In / Out")),
                         ("assets", self.tr("Assets value")),
                         ("result", self.tr("Total result")),
                         ("profit", self.tr("Profit / Loss")),
                         ("dividend", self.tr("Returns")),
                         ("tax_fee", self.tr("Taxes & Fees"))]
        self._view = parent_view
        self._begin = 0
        self._end = 0
        self._account_id = 0
        self._query = None
        self._ym_delegate = None
        self._float_delegate = None
        QSqlTableModel.__init__(self, parent=parent_view, db=db_connection())

    def setColumnNames(self):
        for column in self._columns:
            self.setHeaderData(self.fieldIndex(column[0]), Qt.Horizontal, column[1])

    def resetDelegates(self):
        for column in self._columns:
            self._view.setItemDelegateForColumn(self.fieldIndex(column[0]), None)

    def configureView(self):
        self._view.setModel(self)
        self.setColumnNames()
        self.resetDelegates()
        font = self._view.horizontalHeader().font()
        font.setBold(True)
        self._view.horizontalHeader().setFont(font)
        self._view.setColumnWidth(self.fieldIndex("period"),
                                  self._view.fontMetrics().horizontalAdvance("00/00/0000 00:00:00") * 1.1)
        self._ym_delegate = TimestampDelegate(display_format='%Y %B')
        self._view.setItemDelegateForColumn(self.fieldIndex("period"), self._ym_delegate)
        self._float_delegate = FloatDelegate(2, allow_tail=False)
        self._view.setItemDelegateForColumn(self.fieldIndex("transfer"), self._float_delegate)
        self._view.setItemDelegateForColumn(self.fieldIndex("assets"), self._float_delegate)
        self._view.setItemDelegateForColumn(self.fieldIndex("result"), self._float_delegate)
        self._view.setItemDelegateForColumn(self.fieldIndex("profit"), self._float_delegate)
        self._view.setItemDelegateForColumn(self.fieldIndex("dividend"), self._float_delegate)
        self._view.setItemDelegateForColumn(self.fieldIndex("tax_fee"), self._float_delegate)

    def setDatesRange(self, begin, end):
        self._begin = begin
        self._end = end
        self.calculateProfitLossReport()
        self.configureView()

    def setAccount(self, account_id):
        self._account_id = account_id
        self.calculateProfitLossReport()
        self.configureView()

    def calculateProfitLossReport(self):
        if self._account_id == 0:
            return
        self._query = executeSQL(
            "WITH "
            "_months AS ("
            "SELECT DISTINCT(l.asset_id) AS asset_id, m.m_start AS month, MAX(q.timestamp) AS last_timestamp "
            "FROM ledger AS l "
            "LEFT JOIN "
            "(WITH RECURSIVE months(m_start) AS "
            "( "
            "  VALUES(CAST(strftime('%s', date(:begin, 'unixepoch', 'start of month')) AS INTEGER)) "
            "  UNION ALL "
            "  SELECT CAST(strftime('%s', date(m_start, 'unixepoch', '+1 month')) AS INTEGER) "
            "  FROM months "
            "  WHERE m_start < :end "
            ") "
            "SELECT m_start FROM months) AS m "
            "LEFT JOIN accounts AS a ON l.account_id=a.id "
            "LEFT JOIN quotes AS q ON q.timestamp<=m.m_start AND q.asset_id=l.asset_id AND q.currency_id=a.currency_id "
            "WHERE l.timestamp>=:begin AND l.timestamp<=:end AND l.account_id=:account_id "
            "GROUP BY m.m_start, l.asset_id "
            "ORDER BY m.m_start, l.asset_id ) "
            "SELECT DISTINCT(m.month) AS period, coalesce(t.transfer, 0) AS transfer, coalesce(a.assets, 0) AS assets, "
            "coalesce(p.result, 0) AS result, coalesce(o.profit, 0) AS profit, coalesce(d.dividend, 0) AS dividend, "
            "coalesce(f.tax_fee, 0) AS tax_fee "
            "FROM _months AS m "
            "LEFT JOIN ( "
            "  SELECT mt.month, SUM(-l.amount) AS transfer "
            "  FROM _months AS mt "
            "  LEFT JOIN ledger AS l ON mt.month = "
            "  CAST(strftime('%s', date(l.timestamp, 'unixepoch', 'start of month')) AS INTEGER) "
            "  AND mt.asset_id=l.asset_id "
            "  WHERE l.book_account=:book_transfers AND l.account_id=:account_id GROUP BY mt.month "
            ") AS t ON t.month = m.month "
            "LEFT JOIN ( "
            "  SELECT ma.month, SUM(l.amount*q.quote) AS assets "
            "  FROM _months AS ma "
            "  LEFT JOIN ledger AS l ON l.timestamp<=ma.month AND l.asset_id=ma.asset_id "
            "  LEFT JOIN accounts AS a ON l.account_id=a.id "
            "  LEFT JOIN quotes AS q ON ma.last_timestamp=q.timestamp AND ma.asset_id=q.asset_id AND q.currency_id=a.currency_id "
            "  WHERE l.account_id =:account_id AND (l.book_account=:book_money OR l.book_account=:book_assets) "
            "  GROUP BY ma.month "
            ") AS a ON a.month = m.month "
            "LEFT JOIN ( "
            "  SELECT CAST(strftime('%s', date(l.timestamp, 'unixepoch', 'start of month')) AS INTEGER) AS month,"
            "  SUM(-l.amount) as result"
            "  FROM ledger AS l  "
            "  WHERE (l.book_account=:book_costs OR l.book_account=:book_incomes) AND l.account_id=:account_id "
            "  GROUP BY month "
            ") AS p ON p.month = m.month "
            "LEFT JOIN ( "
            "  SELECT CAST(strftime('%s', date(l.timestamp, 'unixepoch', 'start of month')) "
            "  AS INTEGER) AS month, SUM(-l.amount) as profit "
            "  FROM ledger AS l "
            "  WHERE (l.book_account=:book_costs OR l.book_account=:book_incomes) "
            "  AND category_id=:category_profit AND l.account_id=:account_id "
            "  GROUP BY month "
            ") AS o ON o.month = m.month "
            "LEFT JOIN ( "
            "  SELECT CAST(strftime('%s', date(l.timestamp, 'unixepoch', 'start of month')) AS INTEGER) "
            "  AS month, SUM(-l.amount) as dividend "
            "  FROM ledger AS l "
            "  WHERE (l.book_account=:book_costs OR l.book_account=:book_incomes) "
            "  AND (l.category_id=:category_dividend OR l.category_id=:category_interest) AND l.account_id=:account_id "
            "  GROUP BY month "
            ") AS d ON d.month = m.month "
            "LEFT JOIN ( "
            "  SELECT CAST(strftime('%s', date(l.timestamp, 'unixepoch', 'start of month')) "
            "  AS INTEGER) AS month, SUM(-l.amount) as tax_fee "
            "  FROM ledger AS l "
            "  WHERE l.book_account=:book_costs AND l.category_id<>:category_dividend "
            "AND l.category_id<>:category_interest AND l.account_id=:account_id "
            "  GROUP BY month "
            ") AS f ON f.month = m.month",
            [(":account_id", self._account_id), (":begin", self._begin), (":end", self._end),
             (":book_costs", BookAccount.Costs), (":book_incomes", BookAccount.Incomes),
             (":book_money", BookAccount.Money), (":book_assets", BookAccount.Assets),
             (":book_transfers", BookAccount.Transfers), (":category_profit", PredefinedCategory.Profit),
             (":category_dividend", PredefinedCategory.Dividends), (":category_interest", PredefinedCategory.Interest)],
            forward_only=False)
        self.setQuery(self._query)
        self.modelReset.emit()


# ----------------------------------------------------------------------------------------------------------------------
class ProfitLossReport(QObject):
    def __init__(self):
        super().__init__()
        self.name = self.name = self.tr("P&L by Account")
        self.window_class = "ProfitLossReportWindow"


# ----------------------------------------------------------------------------------------------------------------------
class ProfitLossReportWindow(MdiWidget, Ui_ProfitLossReportWidget):
    def __init__(self, parent=None):
        MdiWidget.__init__(self, parent)
        self.setupUi(self)
        self.parent_mdi = parent

        self.category_model = ProfitLossReportModel(self.ReportTableView)
        self.ReportTableView.setModel(self.category_model)

        self.connect_signals_and_slots()

    def connect_signals_and_slots(self):
        self.ReportRange.changed.connect(self.ReportTableView.model().setDatesRange)
        self.ReportAccountBtn.changed.connect(self.onAccountChange)

    @Slot()
    def onAccountChange(self):
        self.ReportTableView.model().setAccount(self.ReportAccountBtn.account_id)



# This file is obsolete - it was used for Quik HTML reports import for KIT and Uralsib brokers
# But it was created before intermediate JSON statement validation was implememnted.
# So it should be refactored to be a descendant of Statement class

import logging
import math
import re
import pandas
from datetime import datetime, timezone
from PySide6.QtWidgets import QApplication
from jal.db.db import JalDB
from jal.db.operations import Dividend


# -----------------------------------------------------------------------------------------------------------------------
# Strip white spaces from numbers imported form Quik html-report
def convert_amount(val):
    val = val.replace(' ', '')
    try:
        res = float(val)
    except ValueError:
        res = 0
    return res


# -----------------------------------------------------------------------------------------------------------------------
class Quik:
    ClientPattern = "^Код клиента: (.*)$"
    DateTime = 'Дата и время заключения сделки'
    TradeNumber = 'Номер сделки'
    Symbol = 'Код инструмента'
    Name = 'Краткое наименование инструмента'
    Type = 'Направление'
    Qty = 'Кол-во'
    Price = 'Цена'
    Amount = 'Объём'
    Coupon = 'НКД'
    SettleDate = 'Дата расчётов'
    Buy = 'Купля'
    Sell = 'Продажа'
    Fee = 'Комиссия Брокера'
    FeeEx = 'Суммарная комиссия ТС'    # This line is used in KIT Broker reports
    FeeEx1 = 'Комиссия за ИТС'         # Below 3 lines are used in Uralsib Borker reports
    FeeEx2 = 'Комиссия за организацию торговли'
    FeeEx3 = 'Клиринговая комиссия'
    Total = 'ИТОГО'

    def __init__(self, filename):
        self._filename = filename

    def tr(self, text):
        return QApplication.translate("Quik", text)

    def load(self):
        try:
            data = pandas.read_html(self._filename, encoding='cp1251',
                                    converters={self.Qty: convert_amount, self.Amount: convert_amount,
                                                self.Price: convert_amount, self.Coupon: convert_amount})
        except:
            logging.error(self.tr("Can't read statement file"))
            return False

        report_info = data[0]
        deals_info = data[1]
        parts = re.match(self.ClientPattern, report_info[0][2])
        if parts:
            account_id = JalDB().get_account_id(parts.group(1))
        else:
            logging.error(self.tr("Can't get account number from the statement."))
            return False
        if account_id is None:
            logging.error(self.tr("Account with number ") + f"{parts.group(1)}" +
                          self.tr(" not found. Import cancelled."))
            return False

        for index, row in deals_info.iterrows():
            if row[self.Type] == self.Buy:
                qty = int(row[self.Qty])
            elif row[self.Type] == self.Sell:
                qty = -int(row[self.Qty])
            elif row[self.Type][:len(self.Total)] == self.Total:
                break  # End of statement reached
            else:
                logging.warning(self.tr("Unknown operation type ") + f"'{row[self.Type]}'")
                continue
            asset_id = JalDB().get_asset_id(row[self.Symbol])
            if asset_id is None:
                logging.warning(self.tr("Unknown asset ") + f"'{row[self.Symbol]}'")
                continue
            timestamp = int(
                datetime.strptime(row[self.DateTime], "%d.%m.%Y %H:%M:%S").replace(tzinfo=timezone.utc).timestamp())
            settlement = int(
                datetime.strptime(row[self.SettleDate], "%d.%m.%Y").replace(tzinfo=timezone.utc).timestamp())
            number = row[self.TradeNumber]
            price = row[self.Price]
            amount = row[self.Amount]
            lot_size = math.pow(10, round(math.log10(amount / (price * abs(qty)))))
            qty = qty * lot_size
            fee = float(row[self.Fee])
            if self.FeeEx in row:  # Broker dependent fee import
                fee = fee + float(row[self.FeeEx])
            else:
                fee = fee + float(row[self.FeeEx1]) + float(row[self.FeeEx2]) + float(row[self.FeeEx3])
            bond_interest = float(row[self.Coupon])
            JalDB().add_trade(account_id, asset_id, timestamp, settlement, number, qty, price, fee)
            if bond_interest != 0:
                JalDB().add_dividend(Dividend.BondInterest, timestamp, account_id, asset_id,
                                     bond_interest, "НКД", number)
        return True



import logging
from datetime import datetime, timezone

from PySide6.QtWidgets import QApplication
from jal.constants import Setup, PredefinedAsset, PredefinedCategory
from jal.db.helpers import executeSQL, readSQLrecord, readSQL
from jal.db.operations import LedgerTransaction, Dividend, CorporateAction
from jal.db.settings import JalSettings


# -----------------------------------------------------------------------------------------------------------------------
class TaxesRus:
    BOND_PRINCIPAL = 1000  # TODO Principal should be used from 'asset_data' table

    CorpActionText = {
        CorporateAction.SymbolChange: "Смена символа {before} {old} -> {after} {new}",
        CorporateAction.Split: "Сплит {old} {before} в {after}",
        CorporateAction.SpinOff: "Выделение компании {after} {new} из {before:.6f} {old}; доля выделяемого актива {ratio:.2f}%",
        CorporateAction.Merger: "Слияние компании, конвертация {before} {old} в {after} {new}"
    }

    def __init__(self):
        self.account_id = 0
        self.year_begin = 0
        self.year_end = 0
        self.account_currency = ''
        self.account_number = ''
        self.broker_name = ''
        self.broker_iso_cc = "000"
        self.use_settlement = True
        self.reports = {
            "Дивиденды": self.prepare_dividends,
            "Акции": self.prepare_stocks_and_etf,
            "Облигации": self.prepare_bonds,
            "ПФИ": self.prepare_derivatives,
            "Корп.события": self.prepare_corporate_actions,
            "Комиссии": self.prepare_broker_fees,
            "Проценты": self.prepare_broker_interest
        }

    def tr(self, text):
        return QApplication.translate("TaxesRus", text)

    def prepare_tax_report(self, year, account_id, **kwargs):
        tax_report = {}
        self.account_id = account_id
        self.account_number, self.account_currency = \
            readSQL("SELECT a.number, c.symbol FROM accounts AS a "
                    "LEFT JOIN currencies c ON a.currency_id = c.id WHERE a.id=:account",
                    [(":account", account_id)])
        self.year_begin = int(datetime.strptime(f"{year}", "%Y").replace(tzinfo=timezone.utc).timestamp())
        self.year_end = int(datetime.strptime(f"{year + 1}", "%Y").replace(tzinfo=timezone.utc).timestamp())
        self.broker_name, self.broker_iso_cc = readSQL("SELECT b.name AS broker_name, c.iso_code AS country_iso_code "
                                                       "FROM accounts AS a "
                                                       "LEFT JOIN agents AS b ON a.organization_id = b.id "
                                                       "LEFT JOIN countries AS c ON a.country_id = c.id "
                                                       "WHERE a.id=:account", [(":account", account_id)])
        if 'use_settlement' in kwargs:
            self.use_settlement = kwargs['use_settlement']

        self.prepare_exchange_rate_dates()
        for report in self.reports:
            tax_report[report] = self.reports[report]()

        return tax_report

    # Exchange rates are present in database not for every date (and not every possible timestamp)
    # As any action has exact timestamp it won't match rough timestamp of exchange rate most probably
    # Function fills 't_last_dates' table with correspondence between 'real' timestamp and nearest 'exchange' timestamp
    def prepare_exchange_rate_dates(self):
        _ = executeSQL("DELETE FROM t_last_dates")
        _ = executeSQL("INSERT INTO t_last_dates(ref_id, timestamp) "
                       "SELECT ref_id, coalesce(MAX(q.timestamp), 0) AS timestamp "
                       "FROM ("
                       "SELECT d.timestamp AS ref_id FROM dividends AS d WHERE d.account_id = :account_id "
                       "UNION "
                       "SELECT a.timestamp AS ref_id FROM actions AS a WHERE a.account_id = :account_id "
                       "UNION "
                       "SELECT d.open_timestamp AS ref_id FROM deals AS d WHERE d.account_id=:account_id "
                       "UNION "
                       "SELECT d.close_timestamp AS ref_id FROM deals AS d WHERE d.account_id=:account_id "
                       "UNION "
                       "SELECT c.settlement AS ref_id FROM deals AS d LEFT JOIN trades AS c ON "
                       "(c.id=d.open_op_id AND c.op_type=d.open_op_type) OR (c.id=d.close_op_id AND c.op_type=d.close_op_type) "
                       "WHERE d.account_id = :account_id) "
                       "LEFT JOIN accounts AS a ON a.id = :account_id "
                       "LEFT JOIN quotes AS q ON ref_id >= q.timestamp "
                       "AND a.currency_id=q.asset_id AND q.currency_id=:base_currency "
                       "WHERE ref_id IS NOT NULL "    
                       "GROUP BY ref_id", [(":account_id", self.account_id),
                                           (":base_currency", JalSettings().getValue('BaseCurrency'))], commit=True)

    # ------------------------------------------------------------------------------------------------------------------
    # Create a totals row from provided list of dictionaries
    # it calculates sum for each field in fields and adds it to return dictionary
    def insert_totals(self, list_of_values, fields):
        if not list_of_values:
            return
        totals = {"report_template": "totals"}
        for field in fields:
            totals[field] = sum([x[field] for x in list_of_values if field in x])
        list_of_values.append(totals)

    def prepare_dividends(self):
        dividends = []
        query = executeSQL("SELECT d.type, d.timestamp AS payment_date, s.symbol, s.full_name AS full_name, "
                           "s.isin AS isin, d.amount AS amount, d.tax AS tax, q.quote AS rate, p.quote AS price, "
                           "c.name AS country, c.iso_code AS country_iso, c.tax_treaty AS tax_treaty "
                           "FROM dividends AS d "
                           "LEFT JOIN accounts AS a ON d.account_id = a.id "
                           "LEFT JOIN assets_ext AS s ON s.id = d.asset_id AND s.currency_id=a.currency_id "
                           "LEFT JOIN countries AS c ON s.country_id = c.id "
                           "LEFT JOIN t_last_dates AS ld ON d.timestamp=ld.ref_id "
                           "LEFT JOIN quotes AS q ON ld.timestamp=q.timestamp AND a.currency_id=q.asset_id AND q.currency_id=:base_currency "
                           "LEFT JOIN quotes AS p ON d.timestamp=p.timestamp AND d.asset_id=p.asset_id AND p.currency_id=a.currency_id "                           
                           "WHERE d.timestamp>=:begin AND d.timestamp<:end AND d.account_id=:account_id "
                           " AND d.amount>0 AND (d.type=:type_dividend OR d.type=:type_stock_dividend) "
                           "ORDER BY d.timestamp",
                           [(":begin", self.year_begin), (":end", self.year_end), (":account_id", self.account_id),
                            (":base_currency", JalSettings().getValue('BaseCurrency')),
                            (":type_dividend", Dividend.Dividend), (":type_stock_dividend", Dividend.StockDividend)])
        while query.next():
            dividend = readSQLrecord(query, named=True)
            dividend["note"] = ''
            if dividend["type"] == Dividend.StockDividend:
                if not dividend["price"]:
                    logging.error(self.tr("No price data for stock dividend: ") + f"{dividend}")
                    continue
                dividend["amount"] = dividend["amount"] * dividend["price"]
                dividend["note"] = "Дивиденд выплачен в натуральной форме (ценными бумагами)"
            dividend["amount_rub"] = round(dividend["amount"] * dividend["rate"], 2) if dividend["rate"] else 0
            dividend["tax_rub"] = round(dividend["tax"] * dividend["rate"], 2) if dividend["rate"] else 0
            dividend["tax2pay"] = round(0.13 * dividend["amount_rub"], 2)
            if dividend["tax_treaty"]:
                if dividend["tax2pay"] > dividend["tax_rub"]:
                    dividend["tax2pay"] = dividend["tax2pay"] - dividend["tax_rub"]
                else:
                    dividend["tax2pay"] = 0
            dividend['tax_treaty'] = "Да" if dividend['tax_treaty'] else "Нет"
            dividend['report_template'] = "dividend"
            del dividend['type']
            del dividend['price']
            dividends.append(dividend)
        self.insert_totals(dividends, ["amount", "amount_rub", "tax", "tax_rub", "tax2pay"])
        return dividends

    # -----------------------------------------------------------------------------------------------------------------------
    def prepare_stocks_and_etf(self):
        deals = []
        # Take all actions without conversion
        query = executeSQL("SELECT s.symbol AS symbol, s.isin AS isin, d.qty AS qty, cc.iso_code AS country_iso, "
                           "o.timestamp AS o_date, qo.quote AS o_rate, o.settlement AS os_date, o.number AS o_number, "
                           "qos.quote AS os_rate, o.price AS o_price, o.qty AS o_qty, o.fee AS o_fee, "
                           "c.timestamp AS c_date, qc.quote AS c_rate, c.settlement AS cs_date, c.number AS c_number, "
                           "qcs.quote AS cs_rate, c.price AS c_price, c.qty AS c_qty, c.fee AS c_fee, "
                           "SUM(coalesce(-sd.amount*qsd.quote, 0)) AS s_dividend "  # Dividend paid for short position
                           "FROM deals AS d "
                           "JOIN trades AS o ON o.id=d.open_op_id AND o.op_type=d.open_op_type "
                           "JOIN trades AS c ON c.id=d.close_op_id AND c.op_type=d.close_op_type "
                           "LEFT JOIN accounts AS a ON a.id = :account_id "
                           "LEFT JOIN assets_ext AS s ON s.id = o.asset_id AND s.currency_id=a.currency_id "
                           "LEFT JOIN countries AS cc ON cc.id = a.country_id "
                           "LEFT JOIN t_last_dates AS ldo ON o.timestamp=ldo.ref_id "
                           "LEFT JOIN quotes AS qo ON ldo.timestamp=qo.timestamp AND a.currency_id=qo.asset_id AND qo.currency_id=:base_currency "
                           "LEFT JOIN t_last_dates AS ldos ON o.settlement=ldos.ref_id "
                           "LEFT JOIN quotes AS qos ON ldos.timestamp=qos.timestamp AND a.currency_id=qos.asset_id AND qos.currency_id=:base_currency "
                           "LEFT JOIN t_last_dates AS ldc ON c.timestamp=ldc.ref_id "
                           "LEFT JOIN quotes AS qc ON ldc.timestamp=qc.timestamp AND a.currency_id=qc.asset_id AND qc.currency_id=:base_currency "
                           "LEFT JOIN t_last_dates AS ldcs ON c.settlement=ldcs.ref_id "
                           "LEFT JOIN quotes AS qcs ON ldcs.timestamp=qcs.timestamp AND a.currency_id=qcs.asset_id AND qcs.currency_id=:base_currency "
                           "LEFT JOIN dividends AS sd ON d.asset_id=sd.asset_id AND sd.amount<0 "  # Include dividends paid from short positions
                           "AND sd.ex_date>=o_date AND sd.ex_date<=c_date "
                           "LEFT JOIN t_last_dates AS ldsd ON sd.timestamp=ldsd.ref_id "
                           "LEFT JOIN quotes AS qsd ON ldsd.timestamp=qsd.timestamp AND a.currency_id=qsd.asset_id AND qsd.currency_id=:base_currency "
                           "WHERE c.settlement>=:begin AND c.settlement<:end AND d.account_id=:account_id "
                           "AND (s.type_id = :stock OR s.type_id = :fund) "
                           "GROUP BY d.rowid "  # to prevent collapse to 1 line if 'sd' values are NULL
                           "ORDER BY s.symbol, o.timestamp, c.timestamp",
                           [(":begin", self.year_begin), (":end", self.year_end), (":account_id", self.account_id),
                            (":base_currency", JalSettings().getValue('BaseCurrency')),
                            (":stock", PredefinedAsset.Stock), (":fund", PredefinedAsset.ETF)])
        while query.next():
            deal = readSQLrecord(query, named=True)
            if not deal['symbol']:   # there will be row of NULLs if no deals are present (due to SUM aggregation)
                continue
            if not self.use_settlement:
                deal['os_rate'] = deal['o_rate']
                deal['cs_rate'] = deal['c_rate']
            deal['o_type'] = "Покупка" if deal['qty'] >= 0 else "Продажа"
            deal['c_type'] = "Продажа" if deal['qty'] >= 0 else "Покупка"
            deal['o_amount'] = round(deal['o_price'] * abs(deal['qty']), 2)
            deal['o_amount_rub'] = round(deal['o_amount'] * deal['os_rate'], 2) if deal['os_rate'] else 0
            deal['c_amount'] = round(deal['c_price'] * abs(deal['qty']), 2)
            deal['c_amount_rub'] = round(deal['c_amount'] * deal['cs_rate'], 2) if deal['cs_rate'] else 0
            deal['o_fee'] = deal['o_fee'] * abs(deal['qty'] / deal['o_qty'])
            deal['c_fee'] = deal['c_fee'] * abs(deal['qty'] / deal['c_qty'])
            deal['o_fee_rub'] = round(deal['o_fee'] * deal['o_rate'], 2) if deal['o_rate'] else 0
            deal['c_fee_rub'] = round(deal['c_fee'] * deal['c_rate'], 2) if deal['c_rate'] else 0
            deal['income_rub'] = deal['c_amount_rub'] if deal['qty'] >= 0 else deal['o_amount_rub']
            deal['income'] = deal['c_amount'] if deal['qty'] >= 0 else deal['o_amount']
            deal['spending_rub'] = deal['o_amount_rub'] if deal['qty'] >= 0 else deal['c_amount_rub']
            deal['spending_rub'] = deal['spending_rub'] + deal['o_fee_rub'] + deal['c_fee_rub'] + deal['s_dividend']
            deal['spending'] = deal['o_amount'] if deal['qty'] >= 0 else deal['c_amount']
            deal['spending'] = deal['spending'] + deal['o_fee'] + deal['c_fee']
            deal['profit_rub'] = deal['income_rub'] - deal['spending_rub']
            deal['profit'] = deal['income'] - deal['spending']
            if deal['s_dividend'] > 0:  # Dividend was paid during short position
                deal['s_dividend_note'] = f"Удержанный дивиденд: {deal['s_dividend']:.2f} RUB"
            else:
                deal['s_dividend_note'] = ''
            deal['report_template'] = "trade"
            deals.append(deal)
        self.insert_totals(deals, ["income_rub", "spending_rub", "profit_rub", "profit"])
        return deals

    # -----------------------------------------------------------------------------------------------------------------------
    def prepare_bonds(self):
        bonds = []
        # First put all closed deals with bonds
        query = executeSQL("SELECT s.symbol AS symbol, s.isin AS isin, d.qty AS qty, cc.iso_code AS country_iso, "
                           "o.timestamp AS o_date, qo.quote AS o_rate, o.settlement AS os_date, o.number AS o_number, "
                           "qos.quote AS os_rate, o.price AS o_price, o.qty AS o_qty, o.fee AS o_fee, -oi.amount AS o_int, "
                           "c.timestamp AS c_date, qc.quote AS c_rate, c.settlement AS cs_date, c.number AS c_number, "
                           "qcs.quote AS cs_rate, c.price AS c_price, c.qty AS c_qty, c.fee AS c_fee, ci.amount AS c_int "
                           "FROM deals AS d "
                           "JOIN trades AS o ON o.id=d.open_op_id AND o.op_type=d.open_op_type "
                           "LEFT JOIN dividends AS oi ON oi.account_id=:account_id AND oi.number=o.number AND oi.timestamp=o.timestamp AND oi.asset_id=o.asset_id "
                           "JOIN trades AS c ON c.id=d.close_op_id AND c.op_type=d.close_op_type "
                           "LEFT JOIN dividends AS ci ON ci.account_id=:account_id AND ci.number=c.number AND ci.timestamp=c.timestamp AND ci.asset_id=c.asset_id "
                           "LEFT JOIN accounts AS a ON a.id = :account_id "
                           "LEFT JOIN assets_ext AS s ON s.id = o.asset_id AND s.currency_id=a.currency_id "
                           "LEFT JOIN countries AS cc ON cc.id = a.country_id "
                           "LEFT JOIN t_last_dates AS ldo ON o.timestamp=ldo.ref_id "
                           "LEFT JOIN quotes AS qo ON ldo.timestamp=qo.timestamp AND a.currency_id=qo.asset_id AND qo.currency_id=:base_currency "
                           "LEFT JOIN t_last_dates AS ldos ON o.settlement=ldos.ref_id "
                           "LEFT JOIN quotes AS qos ON ldos.timestamp=qos.timestamp AND a.currency_id=qos.asset_id AND qos.currency_id=:base_currency "
                           "LEFT JOIN t_last_dates AS ldc ON c.timestamp=ldc.ref_id "
                           "LEFT JOIN quotes AS qc ON ldc.timestamp=qc.timestamp AND a.currency_id=qc.asset_id AND qc.currency_id=:base_currency "
                           "LEFT JOIN t_last_dates AS ldcs ON c.settlement=ldcs.ref_id "
                           "LEFT JOIN quotes AS qcs ON ldcs.timestamp=qcs.timestamp AND a.currency_id=qcs.asset_id AND qcs.currency_id=:base_currency "
                           "WHERE c.settlement>=:begin AND c.settlement<:end AND d.account_id=:account_id "
                           "AND s.type_id = :bond "
                           "ORDER BY s.symbol, o.timestamp, c.timestamp",
                           [(":begin", self.year_begin), (":end", self.year_end), (":account_id", self.account_id),
                            (":base_currency", JalSettings().getValue('BaseCurrency')),
                            (":bond", PredefinedAsset.Bond)])
        while query.next():
            deal = readSQLrecord(query, named=True)
            deal['principal'] = self.BOND_PRINCIPAL
            if not self.use_settlement:
                deal['os_rate'] = deal['o_rate']
                deal['cs_rate'] = deal['c_rate']
            deal['o_type'] = "Покупка" if deal['qty'] >= 0 else "Продажа"
            deal['c_type'] = "Продажа" if deal['qty'] >= 0 else "Покупка"
            deal['o_amount'] = round(deal['o_price'] * abs(deal['qty']), 2)
            deal['o_amount_rub'] = round(deal['o_amount'] * deal['os_rate'], 2) if deal['os_rate'] else 0
            deal['c_amount'] = round(deal['c_price'] * abs(deal['qty']), 2)
            deal['c_amount_rub'] = round(deal['c_amount'] * deal['cs_rate'], 2) if deal['cs_rate'] else 0
            # Convert price from currency to % of principal
            deal['o_price'] = 100.0 * deal['o_price'] / deal['principal']
            deal['c_price'] = 100.0 * deal['c_price'] / deal['principal']

            deal['o_fee'] = deal['o_fee'] * abs(deal['qty'] / deal['o_qty'])
            deal['c_fee'] = deal['c_fee'] * abs(deal['qty'] / deal['c_qty'])
            deal['o_fee_rub'] = round(deal['o_fee'] * deal['o_rate'], 2) if deal['o_rate'] else 0
            deal['c_fee_rub'] = round(deal['c_fee'] * deal['c_rate'], 2) if deal['c_rate'] else 0
            deal['o_int_rub'] = round(deal['o_int'] * deal['o_rate'], 2) if deal['o_rate'] and deal['o_int'] else 0
            deal['c_int_rub'] = round(deal['c_int'] * deal['o_rate'], 2) if deal['o_rate'] and deal['c_int'] else 0
            # TODO accrued interest calculations for short deals is not clear - to be corrected
            deal['income_rub'] = deal['c_amount_rub'] + deal['c_int_rub'] if deal['qty'] >= 0 else deal['o_amount_rub']
            deal['income'] = deal['c_amount'] if deal['qty'] >= 0 else deal['o_amount']
            deal['spending_rub'] = deal['o_amount_rub'] if deal['qty'] >= 0 else deal['c_amount_rub']
            deal['spending_rub'] = deal['spending_rub'] + deal['o_fee_rub'] + deal['c_fee_rub'] + deal['o_int_rub']
            deal['spending'] = deal['o_amount'] if deal['qty'] >= 0 else deal['c_amount']
            deal['spending'] = deal['spending'] + deal['o_fee'] + deal['c_fee']
            deal['profit_rub'] = deal['income_rub'] - deal['spending_rub']
            deal['profit'] = deal['income'] - deal['spending']
            deal['report_template'] = "bond_trade"
            bonds.append(deal)

        # Second - take all bond interest payments not linked with buy/sell transactions
        query = executeSQL("SELECT b.symbol AS symbol, b.isin AS isin, i.timestamp AS o_date, i.number AS number, "
                           "i.amount AS interest, r.quote AS rate, cc.iso_code AS country_iso "
                           "FROM dividends AS i "
                           "LEFT JOIN trades AS t ON i.account_id=t.account_id AND i.number=t.number "
                           "AND i.timestamp=t.timestamp AND i.asset_id=t.asset_id "
                           "LEFT JOIN accounts AS a ON a.id = i.account_id "
                           "LEFT JOIN assets_ext AS b ON b.id = i.asset_id AND b.currency_id=a.currency_id "
                           "LEFT JOIN countries AS cc ON cc.id = a.country_id "
                           "LEFT JOIN t_last_dates AS ld ON i.timestamp=ld.ref_id "
                           "LEFT JOIN quotes AS r ON ld.timestamp=r.timestamp AND a.currency_id=r.asset_id AND r.currency_id=:base_currency "
                           "WHERE i.timestamp>=:begin AND i.timestamp<:end AND i.account_id=:account_id "
                           "AND i.type = :type_interest AND t.id IS NULL",
                           [(":begin", self.year_begin), (":end", self.year_end), (":account_id", self.account_id),
                            (":base_currency", JalSettings().getValue('BaseCurrency')),
                            (":type_interest", Dividend.BondInterest)])
        while query.next():
            interest = readSQLrecord(query, named=True)
            interest['type'] = "Купон"
            interest['empty'] = ''  # to keep cell borders drawn
            interest['interest_rub'] = round(interest['interest'] * interest['rate'], 2) if interest['rate'] else 0
            interest['income_rub'] = interest['profit_rub'] = interest['interest_rub']
            interest['spending_rub'] = 0.0
            interest['profit'] = interest['interest']
            interest['report_template'] = "bond_interest"
            bonds.append(interest)
        self.insert_totals(bonds, ["income_rub", "spending_rub", "profit_rub", "profit"])
        return bonds

    # -----------------------------------------------------------------------------------------------------------------------
    def prepare_derivatives(self):
        derivatives = []
        # Take all actions without conversion
        query = executeSQL("SELECT s.symbol, d.qty AS qty, cc.iso_code AS country_iso, "
                           "o.timestamp AS o_date, qo.quote AS o_rate, o.settlement AS os_date, o.number AS o_number, "
                           "qos.quote AS os_rate, o.price AS o_price, o.qty AS o_qty, o.fee AS o_fee, "
                           "c.timestamp AS c_date, qc.quote AS c_rate, c.settlement AS cs_date, c.number AS c_number, "
                           "qcs.quote AS cs_rate, c.price AS c_price, c.qty AS c_qty, c.fee AS c_fee "
                           "FROM deals AS d "
                           "JOIN trades AS o ON o.id=d.open_op_id AND o.op_type=d.open_op_type "
                           "JOIN trades AS c ON c.id=d.close_op_id AND c.op_type=d.close_op_type "
                           "LEFT JOIN accounts AS a ON a.id = :account_id "
                           "LEFT JOIN assets_ext AS s ON s.id = o.asset_id AND s.currency_id=a.currency_id "
                           "LEFT JOIN countries AS cc ON cc.id = a.country_id "
                           "LEFT JOIN t_last_dates AS ldo ON o.timestamp=ldo.ref_id "
                           "LEFT JOIN quotes AS qo ON ldo.timestamp=qo.timestamp AND a.currency_id=qo.asset_id AND qo.currency_id=:base_currency "
                           "LEFT JOIN t_last_dates AS ldos ON o.settlement=ldos.ref_id "
                           "LEFT JOIN quotes AS qos ON ldos.timestamp=qos.timestamp AND a.currency_id=qos.asset_id AND qos.currency_id=:base_currency "
                           "LEFT JOIN t_last_dates AS ldc ON c.timestamp=ldc.ref_id "
                           "LEFT JOIN quotes AS qc ON ldc.timestamp=qc.timestamp AND a.currency_id=qc.asset_id AND qc.currency_id=:base_currency "
                           "LEFT JOIN t_last_dates AS ldcs ON c.settlement=ldcs.ref_id "
                           "LEFT JOIN quotes AS qcs ON ldcs.timestamp=qcs.timestamp AND a.currency_id=qcs.asset_id AND qcs.currency_id=:base_currency "
                           "WHERE c.settlement>=:begin AND c.settlement<:end AND d.account_id=:account_id "
                           "AND s.type_id = :derivative "
                           "ORDER BY s.symbol, o.timestamp, c.timestamp",
                           [(":begin", self.year_begin), (":end", self.year_end), (":account_id", self.account_id),
                            (":base_currency", JalSettings().getValue('BaseCurrency')),
                            (":derivative", PredefinedAsset.Derivative)])
        while query.next():
            deal = readSQLrecord(query, named=True)
            if not self.use_settlement:
                deal['os_rate'] = deal['o_rate']
                deal['cs_rate'] = deal['c_rate']
            deal['o_type'] = "Покупка" if deal['qty'] >= 0 else "Продажа"
            deal['c_type'] = "Продажа" if deal['qty'] >= 0 else "Покупка"
            deal['o_amount'] = round(deal['o_price'] * abs(deal['qty']), 2)
            deal['o_amount_rub'] = round(deal['o_amount'] * deal['os_rate'], 2) if deal['os_rate'] else 0
            deal['c_amount'] = round(deal['c_price'] * abs(deal['qty']), 2)
            deal['c_amount_rub'] = round(deal['c_amount'] * deal['cs_rate'], 2) if deal['cs_rate'] else 0
            deal['o_fee'] = deal['o_fee'] * abs(deal['qty'] / deal['o_qty'])
            deal['c_fee'] = deal['c_fee'] * abs(deal['qty'] / deal['c_qty'])
            deal['o_fee_rub'] = round(deal['o_fee'] * deal['o_rate'], 2) if deal['o_rate'] else 0
            deal['c_fee_rub'] = round(deal['c_fee'] * deal['c_rate'], 2) if deal['c_rate'] else 0
            deal['income_rub'] = deal['c_amount_rub'] if deal['qty'] >= 0 else deal['o_amount_rub']
            deal['income'] = deal['c_amount'] if deal['qty'] >= 0 else deal['o_amount']
            deal['spending_rub'] = deal['o_amount_rub'] if deal['qty'] >= 0 else deal['c_amount_rub']
            deal['spending_rub'] = deal['spending_rub'] + deal['o_fee_rub'] + deal['c_fee_rub']
            deal['spending'] = deal['o_amount'] if deal['qty'] >= 0 else deal['c_amount']
            deal['spending'] = deal['spending'] + deal['o_fee'] + deal['c_fee']
            deal['profit_rub'] = deal['income_rub'] - deal['spending_rub']
            deal['profit'] = deal['income'] - deal['spending']
            deal['report_template'] = "trade"
            derivatives.append(deal)
        self.insert_totals(derivatives, ["income_rub", "spending_rub", "profit_rub", "profit"])
        return derivatives

    # -----------------------------------------------------------------------------------------------------------------------
    def prepare_broker_fees(self):
        fees = []
        query = executeSQL("SELECT a.timestamp AS payment_date, d.amount AS amount, d.note AS note, q.quote AS rate "
                           "FROM actions AS a "
                           "LEFT JOIN action_details AS d ON d.pid=a.id "
                           "LEFT JOIN accounts AS c ON c.id = :account_id "
                           "LEFT JOIN t_last_dates AS ld ON a.timestamp=ld.ref_id "
                           "LEFT JOIN quotes AS q ON ld.timestamp=q.timestamp AND c.currency_id=q.asset_id AND q.currency_id=:base_currency "
                           "WHERE a.timestamp>=:begin AND a.timestamp<:end "
                           "AND a.account_id=:account_id AND d.category_id=:fee",
                           [(":begin", self.year_begin), (":end", self.year_end),
                            (":account_id", self.account_id), (":fee", PredefinedCategory.Fees),
                            (":base_currency", JalSettings().getValue('BaseCurrency'))])
        while query.next():
            fee = readSQLrecord(query, named=True)
            fee['amount'] = -fee['amount']
            fee['amount_rub'] = round(fee['amount'] * fee['rate'], 2) if fee['rate'] else 0
            fee['report_template'] = "fee"
            fees.append(fee)
        self.insert_totals(fees, ["amount", "amount_rub"])
        return fees

    # -----------------------------------------------------------------------------------------------------------------------
    def prepare_broker_interest(self):
        interests = []
        query = executeSQL("SELECT a.timestamp AS payment_date, d.amount AS amount, d.note AS note, q.quote AS rate "
                           "FROM actions AS a "
                           "LEFT JOIN action_details AS d ON d.pid=a.id "
                           "LEFT JOIN accounts AS c ON c.id = :account_id "
                           "LEFT JOIN t_last_dates AS ld ON a.timestamp=ld.ref_id "
                           "LEFT JOIN quotes AS q ON ld.timestamp=q.timestamp AND c.currency_id=q.asset_id AND q.currency_id=:base_currency "
                           "WHERE a.timestamp>=:begin AND a.timestamp<:end "
                           "AND a.account_id=:account_id AND d.category_id=:fee",
                           [(":begin", self.year_begin), (":end", self.year_end),
                            (":account_id", self.account_id), (":fee", PredefinedCategory.Interest),
                            (":base_currency", JalSettings().getValue('BaseCurrency'))])
        while query.next():
            interest = readSQLrecord(query, named=True)
            interest['amount'] = interest['amount']
            interest['amount_rub'] = round(interest['amount'] * interest['rate'], 2) if interest['rate'] else 0
            interest['tax_rub'] = round(0.13 * interest['amount_rub'], 2)
            interest['report_template'] = "interest"
            interests.append(interest)
        self.insert_totals(interests, ["amount", "amount_rub", "tax_rub"])
        return interests

    # -----------------------------------------------------------------------------------------------------------------------
    def prepare_corporate_actions(self):
        corp_actions = []
        # get list of all deals that were opened with corp.action and closed by normal trade
        query = executeSQL("SELECT d.open_op_id AS operation_id, s.symbol, d.qty AS qty, "
                           "t.number AS trade_number, t.timestamp AS t_date, qt.quote AS t_rate, "
                           "t.settlement AS s_date, qts.quote AS s_rate, t.price AS price, t.fee AS fee, "
                           "s.full_name AS full_name, s.isin AS isin, s.type_id AS type_id "
                           "FROM deals AS d "
                           "JOIN trades AS t ON t.id=d.close_op_id AND t.op_type=d.close_op_type "
                           "LEFT JOIN accounts AS a ON a.id = :account_id "
                           "LEFT JOIN assets_ext AS s ON s.id = t.asset_id AND s.currency_id=a.currency_id "
                           "LEFT JOIN t_last_dates AS ldt ON t.timestamp=ldt.ref_id "
                           "LEFT JOIN quotes AS qt ON ldt.timestamp=qt.timestamp AND a.currency_id=qt.asset_id AND qt.currency_id=:base_currency "
                           "LEFT JOIN t_last_dates AS ldts ON t.settlement=ldts.ref_id "
                           "LEFT JOIN quotes AS qts ON ldts.timestamp=qts.timestamp AND a.currency_id=qts.asset_id AND qts.currency_id=:base_currency "
                           "WHERE t.settlement<:end AND d.account_id=:account_id AND d.open_op_type=:corp_action "
                           "ORDER BY s.symbol, t.timestamp",
                           [(":end", self.year_end), (":account_id", self.account_id),
                            (":corp_action", LedgerTransaction.CorporateAction),
                            (":base_currency", JalSettings().getValue('BaseCurrency'))])
        group = 1
        basis = 1
        previous_symbol = ""
        while query.next():
            actions = []
            sale = readSQLrecord(query, named=True)
            if previous_symbol != sale['symbol']:
                # Clean processed qty records if symbol have changed
                _ = executeSQL("DELETE FROM t_last_assets")
                if sale["s_date"] >= self.year_begin:  # Don't put sub-header of operation is out of scope
                    corp_actions.append({'report_template': "symbol_header",
                                    'report_group': 0,
                                    'description': f"Сделки по бумаге: {sale['symbol']} - {sale['full_name']}"})
                    previous_symbol = sale['symbol']
            sale['operation'] = "Продажа"
            sale['basis_ratio'] = 100.0 * basis
            sale['amount'] = round(sale['price'] * sale['qty'], 2)
            if sale['s_rate']:
                sale['amount_rub'] = round(sale['amount'] * sale['s_rate'], 2)
            else:
                sale['amount_rub'] = 0
            if sale['t_rate']:
                sale['fee_rub'] = round(sale['fee'] * sale['t_rate'], 2)
            else:
                sale['fee_rub'] = 0
            sale['income_rub'] = sale['amount_rub']
            sale['spending_rub'] = sale['fee_rub']

            if sale["t_date"] < self.year_begin:    # Don't show deal that is before report year (level = -1)
                self.proceed_corporate_action(actions, sale['operation_id'], sale['symbol'], sale['qty'], basis, -1, group)
            else:
                sale['report_template'] = "trade"
                sale['report_group'] = group
                actions.append(sale)
                if sale['type_id'] == PredefinedAsset.Bond:
                    self.output_accrued_interest(actions, sale['trade_number'], 1, 0)
                self.proceed_corporate_action(actions, sale['operation_id'], sale['symbol'], sale['qty'], basis, 1, group)

            self.insert_totals(actions, ["income_rub", "spending_rub"])
            corp_actions += actions
            group += 1
        return corp_actions

    def proceed_corporate_action(self, actions, operation_id, symbol, qty, basis, level, group):
        qty, symbol, basis = self.output_corp_action(actions, operation_id, symbol, qty, basis, level, group)
        next_level = -1 if level == -1 else (level + 1)
        self.next_corporate_action(actions, operation_id, symbol, qty, basis, next_level, group)

    # operation_id - id of corporate action
    def next_corporate_action(self, actions, operation_id, symbol, qty, basis, level, group):
        # get list of deals that were closed as result of current corporate action
        open_query = executeSQL("SELECT open_op_id AS open_op_id, open_op_type AS op_type "
                                "FROM deals "
                                "WHERE close_op_id=:close_op_id AND close_op_type=:corp_action "
                                "ORDER BY id",
                                [(":close_op_id", operation_id), (":corp_action", LedgerTransaction.CorporateAction)])
        while open_query.next():
            open_id, open_type = readSQLrecord(open_query)

            if open_type == LedgerTransaction.Trade:
                qty = self.output_purchase(actions, open_id, qty, basis, level, group)
            elif open_type == LedgerTransaction.CorporateAction:
                self.proceed_corporate_action(actions, open_id, symbol, qty, basis, level, group)
            else:
                assert False

    # operation_id - id of buy operation
    def output_purchase(self, actions, operation_id, proceed_qty, basis, level, group):
        if proceed_qty <= 0:
            return proceed_qty

        purchase = readSQL("SELECT t.id AS trade_id, s.symbol, s.isin AS isin, s.type_id AS type_id, "
                           "coalesce(d.qty-SUM(lq.total_value), d.qty) AS qty, "
                           "t.timestamp AS t_date, qt.quote AS t_rate, t.number AS trade_number, "
                           "t.settlement AS s_date, qts.quote AS s_rate, t.price AS price, t.fee AS fee "
                           "FROM trades AS t "
                           "JOIN deals AS d ON t.id=d.open_op_id AND t.op_type=d.open_op_type "
                           "LEFT JOIN accounts AS a ON a.id = t.account_id "
                           "LEFT JOIN assets_ext AS s ON s.id = t.asset_id AND s.currency_id=a.currency_id "
                           "LEFT JOIN t_last_dates AS ldt ON t.timestamp=ldt.ref_id "
                           "LEFT JOIN quotes AS qt ON ldt.timestamp=qt.timestamp AND a.currency_id=qt.asset_id AND qt.currency_id=:base_currency "
                           "LEFT JOIN t_last_dates AS ldts ON t.settlement=ldts.ref_id "
                           "LEFT JOIN quotes AS qts ON ldts.timestamp=qts.timestamp AND a.currency_id=qts.asset_id AND qts.currency_id=:base_currency "
                           "LEFT JOIN t_last_assets AS lq ON lq.id = t.id "
                           "WHERE t.id = :operation_id", [(":operation_id", operation_id),
                                                          (":base_currency", JalSettings().getValue('BaseCurrency'))],
                           named=True)
        if purchase['qty'] <= (2 * Setup.CALC_TOLERANCE):
            return proceed_qty  # This trade was fully mached before

        purchase['operation'] = ' ' * level * 3 + "Покупка"
        purchase['basis_ratio'] = 100.0 * basis
        deal_qty = purchase['qty']
        purchase['qty'] = proceed_qty if proceed_qty < deal_qty else deal_qty
        purchase['amount'] = round(purchase['price'] * purchase['qty'], 2)
        purchase['amount_rub'] = round(purchase['amount'] * purchase['s_rate'], 2) if purchase['s_rate'] else 0
        purchase['fee'] = purchase['fee'] * purchase['qty'] / deal_qty
        purchase['fee_rub'] = round(purchase['fee'] * purchase['t_rate'], 2) if purchase['t_rate'] else 0
        purchase['income_rub'] = 0
        purchase['spending_rub'] = round(basis*(purchase['amount_rub'] + purchase['fee_rub']), 2)

        _ = executeSQL("INSERT INTO t_last_assets (id, total_value) VALUES (:trade_id, :qty)",
                       [(":trade_id", purchase['trade_id']), (":qty", purchase['qty'])])
        if level >= 0:  # Don't output if level==-1, i.e. corp action is out of report scope
            purchase['report_template'] = "trade"
            purchase['report_group'] = group
            actions.append(purchase)
        if purchase['type_id'] == PredefinedAsset.Bond:
            share = purchase['qty'] / deal_qty if purchase['qty'] < deal_qty else 1
            self.output_accrued_interest(actions, purchase['trade_number'], share, level)
        return proceed_qty - purchase['qty']

    def output_corp_action(self, actions, operation_id, symbol, proceed_qty, basis, level, group):
        if proceed_qty <= 0:
            return proceed_qty

        action = readSQL("SELECT c.timestamp AS action_date, c.number AS action_number, c.type, "
                         "s1.symbol AS symbol, s1.isin AS isin, c.qty AS qty, "
                         "s2.symbol AS symbol_new, s2.isin AS isin_new, c.qty_new AS qty_new, "
                         "c.note AS note, c.basis_ratio "
                         "FROM corp_actions  c "
                         "LEFT JOIN accounts a ON c.account_id=a.id "
                         "LEFT JOIN assets_ext  s1 ON c.asset_id=s1.id AND s1.currency_id=a.currency_id "
                         "LEFT JOIN assets_ext  s2 ON c.asset_id_new=s2.id AND s1.currency_id=a.currency_id "
                         "WHERE c.id = :operation_id ",
                         [(":operation_id", operation_id)], named=True)
        action['operation'] = ' ' * level * 3 + "Корп. действие"
        old_asset = f"{action['symbol']} ({action['isin']})"
        new_asset = f"{action['symbol_new']} ({action['isin_new']})"
        if action['type'] == CorporateAction.SpinOff:
            action['description'] = self.CorpActionText[action['type']].format(old=old_asset, new=new_asset,
                                                                               before=action['qty'],
                                                                               after=action['qty_new'],
                                                                               ratio=100.0 * action['basis_ratio'])
            if symbol == action['symbol_new']:
                basis = basis * action['basis_ratio']
                qty_before = action['qty'] * proceed_qty / action['qty_new']
            else:
                basis = basis * (1 - action['basis_ratio'])
                qty_before = action['qty']
        else:
            qty_before = action['qty'] * proceed_qty / action['qty_new']
            qty_after = proceed_qty
            action['description'] = self.CorpActionText[action['type']].format(old=old_asset, new=new_asset,
                                                                               before=qty_before, after=qty_after)
        if level >= 0:  # Don't output if level==-1, i.e. corp action is out of report scope
            action['report_template'] = "action"
            action['report_group'] = group
            actions.append(action)
        return qty_before, action['symbol'], basis

    def output_accrued_interest(self, actions, trade_number, share, level):
        interest = readSQL("SELECT b.symbol AS symbol, b.isin AS isin, i.timestamp AS o_date, i.number AS number, "
                           "i.amount AS interest, r.quote AS rate, cc.iso_code AS country_iso "
                           "FROM dividends AS i "
                           "LEFT JOIN accounts AS a ON a.id = i.account_id "
                           "LEFT JOIN assets_ext AS b ON b.id = i.asset_id AND b.currency_id=a.currency_id "
                           "LEFT JOIN countries AS cc ON cc.id = a.country_id "
                           "LEFT JOIN t_last_dates AS ld ON i.timestamp=ld.ref_id "
                           "LEFT JOIN quotes AS r ON ld.timestamp=r.timestamp AND a.currency_id=r.asset_id AND r.currency_id=:base_currency "
                           "WHERE i.account_id=:account_id AND i.type=:interest AND i.number=:trade_number",
                           [(":account_id", self.account_id), (":interest", Dividend.BondInterest),
                            (":trade_number", trade_number),
                            (":base_currency", JalSettings().getValue('BaseCurrency'))], named=True)
        if interest is None:
            return
        interest['empty'] = ''
        interest['interest'] = interest['interest'] if share == 1 else share * interest['interest']
        interest['interest_rub'] = abs(round(interest['interest'] * interest['rate'], 2)) if interest['rate'] else 0
        if interest['interest'] < 0:  # Accrued interest paid for purchase
            interest['interest'] = -interest['interest']
            interest['operation'] = ' ' * level * 3 + "НКД уплачен"
            interest['spending_rub'] = interest['interest_rub']
            interest['income_rub'] = 0.0
        else:                         # Accrued interest received for sale
            interest['operation'] = ' ' * level * 3 + "НКД получен"
            interest['income_rub'] = interest['interest_rub']
            interest['spending_rub'] = 0.0
        interest['report_template'] = "bond_interest"
        actions.append(interest)



from functools import partial
from datetime import datetime
import logging

from PySide6.QtCore import Property, Slot
from PySide6.QtWidgets import QFileDialog
from jal.ui.ui_tax_export_widget import Ui_TaxWidget
from jal.widgets.mdi import MdiWidget
from jal.data_export.taxes import TaxesRus
from jal.data_export.xlsx import XLSX
from jal.data_export.dlsg import DLSG


class TaxWidget(MdiWidget, Ui_TaxWidget):
    def __init__(self, parent):
        MdiWidget.__init__(self)
        self.setupUi(self)

        self.Year.setValue(datetime.now().year - 1)   # Set previous year by default
        self.XlsSelectBtn.pressed.connect(partial(self.OnFileBtn, 'XLS'))
        self.DlsgSelectBtn.pressed.connect(partial(self.OnFileBtn, 'DLSG'))
        self.SaveButton.pressed.connect(self.SaveReport)

        # center dialog with respect to parent window
        x = parent.x() + parent.width() / 2 - self.width() / 2
        y = parent.y() + parent.height() / 2 - self.height() / 2
        self.setGeometry(x, y, self.width(), self.height())

    @Slot()
    def OnFileBtn(self, type):
        if type == 'XLS':
            selector = (self.tr("Save tax reports to:"), self.tr("Excel files (*.xlsx)"), '.xlsx', self.XlsFileName)
        elif type == 'DLSG':
            last_digit = self.year % 10
            selector = (self.tr("Save tax form to:"), self.tr(f"Tax form (*.dc{last_digit})"),
                        f".dc{last_digit}", self.DlsgFileName)
        else:
            raise ValueError
        filename = QFileDialog.getSaveFileName(self, selector[0], ".", selector[1])
        if filename[0]:
            if filename[1] == selector[1] and filename[0][-len(selector[2]):] != selector[2]:
                selector[3].setText(filename[0] + selector[2])
            else:
                selector[3].setText(filename[0])

    def getYear(self):
        return self.Year.value()

    def getXlsFilename(self):
        return self.XlsFileName.text()

    def getAccount(self):
        return self.AccountWidget.selected_id

    def getDlsgState(self):
        return self.DlsgGroup.isChecked()

    def getDslgFilename(self):
        return self.DlsgFileName.text()

    def getBrokerAsIncomeName(self):
        return self.IncomeSourceBroker.isChecked()

    def getDividendsOnly(self):
        return self.DividendsOnly.isChecked()

    def getNoSettlement(self):
        return self.NoSettlement.isChecked()

    year = Property(int, fget=getYear)
    xls_filename = Property(str, fget=getXlsFilename)
    account = Property(int, fget=getAccount)
    update_dlsg = Property(bool, fget=getDlsgState)
    dlsg_filename = Property(str, fget=getDslgFilename)
    dlsg_broker_as_income = Property(bool, fget=getBrokerAsIncomeName)
    dlsg_dividends_only = Property(bool, fget=getDividendsOnly)
    no_settelement = Property(bool, fget=getNoSettlement)

    def SaveReport(self):
        taxes = TaxesRus()
        tax_report = taxes.prepare_tax_report(self.year, self.account, use_settlement=(not self.no_settelement))

        reports_xls = XLSX(self.xls_filename)
        templates = {
            "Дивиденды": "tax_rus_dividends.json",
            "Акции": "tax_rus_trades.json",
            "Облигации": "tax_rus_bonds.json",
            "ПФИ": "tax_rus_derivatives.json",
            "Корп.события": "tax_rus_corporate_actions.json",
            "Комиссии": "tax_rus_fees.json",
            "Проценты": "tax_rus_interests.json"
        }
        parameters = {
            "period": f"{datetime.utcfromtimestamp(taxes.year_begin).strftime('%d.%m.%Y')}"
                      f" - {datetime.utcfromtimestamp(taxes.year_end - 1).strftime('%d.%m.%Y')}",
            "account": f"{taxes.account_number} ({taxes.account_currency})",
            "currency": taxes.account_currency,
            "broker_name": taxes.broker_name,
            "broker_iso_country": taxes.broker_iso_cc
        }
        for section in tax_report:
            if section not in templates:
                continue
            reports_xls.output_data(tax_report[section], templates[section], parameters)
        reports_xls.save()

        logging.info(self.tr("Tax report saved to file ") + f"'{self.xls_filename}'")

        if self.update_dlsg:
            tax_forms = DLSG(self.year, broker_as_income=self.dlsg_broker_as_income,
                             only_dividends=self.dlsg_dividends_only)
            tax_forms.update_taxes(tax_report, parameters)
            try:
                tax_forms.save(self.dlsg_filename)
            except:
                logging.error(self.tr("Can't write tax form into file ") + f"'{self.dlsg_filename}'")



import logging
from copy import deepcopy
from datetime import date, datetime
from PySide6.QtWidgets import QApplication


# ----------------------------------------------------------------------------------------------------------------------
# Basic 3-ndfl templates derived from dcX files
form_3nfl_template = {
    2020: {
        "header": "DLSG            Decl20200102FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF",
        "sections": {
            "@DeclInfo": ('', '', 0, 0, datetime(2022, 1, 29), 0, 5, 0, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, '', 0, '', '', 0),
            "@PersonName": ('', '', '', '', '', 0),
            "@PersonDocument": (0, '', '', datetime(1990, 1, 1), datetime(1977, 1, 1), '', '', 0),
            "@Foreigner": ('', "РОССИЯ", '', 643),
            "@PhoneForeignerHome": ('', ''),
            "@PhoneForeignerWork": ('', ''),
            "@PersonAddress": (0, '', 0, '', '', '', '', '', '', '', '', ''),
            "@HomePhone": ('', ''),
            "@WorkPhone": ('', ''),
            "@DeclInquiry": (0, 0, 0, 0, 0),
            "@DeclForeign": {
                # "@CurrencyIncome000": (13, 1530, "(01)Доходы от реализации ЦБ (обращ-ся на орг. рынке ЦБ)", "IBKR", 840,
                #                        datetime(2020, 1, 30), datetime(2020, 1, 30), 1, 840, 6239.34, 100, 6239.34, 100,
                #                        "Доллар США", 123, 7674.39, 11, 686.33, 201, 321, 0, 0, '', 0)
            },
            "@StandartDeduct": (0, ),
            "@SocialDeduct": (0, ),
            "@ConstructionDeduct": (0, ),
            "@CBDeduct": (0, ),
            "@InvDeduct": (0, )
        }
    },
    2021: {
        "header": "DLSG            Decl20210103FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF",
        "sections": {
            "@DeclInfo": ('', '', 0, 0, datetime(2022, 1, 29), 0, 5, 0, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, '', 0, '', '', 0),
            "@PersonName": ('', '', '', '', '', 0),
            "@PersonDocument": (0, '', '', datetime(1990, 1, 1), datetime(1977, 1, 1), '', '', 0),
            "@Foreigner": ('', "РОССИЯ", '', 643),
            "@PhoneForeignerHome": ('', ''),
            "@PhoneForeignerWork": ('', ''),
            "@PersonAddress": (0, '', 0, '', '', '', '', '', '', '', '', ''),
            "@HomePhone": ('', ''),
            "@WorkPhone": ('', ''),
            "@DeclInquiry": (0, 0, 0, 0, 0, 0),
            "@DeclForeign": {
                # "@CurrencyIncome0000": (0, 1530, "(01)Доходы от реализации ЦБ (обращ-ся на орг. рынке ЦБ)", "IBKR", 324,
                #     840, datetime(2021, 1, 29), datetime(2021, 1, 29), 1, 840, 7618.54, 100, 7618.54, 100, "Доллар США",
                #     123, 9370.8, 11, 838.04, 201, 321, 0, 0, 0, '', 0,)
            },
            "@StandartDeduct": (0, ),
            "@SocialDeduct": (0, ),
            "@ConstructionDeduct": (0, ),
            "@CBDeduct": (0, ),
            "@InvDeduct": (0, )
        }
    }
}


# ----------------------------------------------------------------------------------------------------------------------
class DLSG:
    currencies = {
        'AUD': {'code': '036', 'name': 'Австралийский доллар', 'multiplier': 100},
        'AZN': {'code': '944', 'name': 'Азербайджанский манат', 'multiplier': 100},
        'GBP': {'code': '826', 'name': 'Фунт стерлингов', 'multiplier': 100},
        'AMD': {'code': '051', 'name': 'Армянский драм', 'multiplier': 10000},
        'BYN': {'code': '933', 'name': 'Белорусский рубль', 'multiplier': 100},
        'BGN': {'code': '975', 'name': 'Болгарский лев', 'multiplier': 100},
        'BRL': {'code': '986', 'name': 'Бразильский реал', 'multiplier': 100},
        'HUF': {'code': '348', 'name': 'Форинт', 'multiplier': 10000},
        'HKD': {'code': '344', 'name': 'Гонконгский доллар', 'multiplier': 1000},
        'DKK': {'code': '208', 'name': 'Датская крона', 'multiplier': 100},
        'USD': {'code': '840', 'name': 'Доллар США', 'multiplier': 100},
        'EUR': {'code': '978', 'name': 'Евро', 'multiplier': 100},
        'INR': {'code': '356', 'name': 'Индийская рупия', 'multiplier': 10000},
        'KZT': {'code': '398', 'name': 'Тенге', 'multiplier': 10000},
        'CAD': {'code': '124', 'name': 'Канадский доллар', 'multiplier': 100},
        'KGS': {'code': '417', 'name': 'Сом', 'multiplier': 10000},
        'CNY': {'code': '156', 'name': 'Юань', 'multiplier': 100},
        'MDL': {'code': '498', 'name': 'Молдавский лей', 'multiplier': 1000},
        'NOK': {'code': '578', 'name': 'Норвежская крона', 'multiplier': 1000},
        'PLN': {'code': '985', 'name': 'Злотый', 'multiplier': 100},
        'RON': {'code': '946', 'name': 'Румынский лей', 'multiplier': 100},
        'SGD': {'code': '702', 'name': 'Сингапурский доллар', 'multiplier': 100},
        'TJS': {'code': '972', 'name': 'Сомони', 'multiplier': 1000},
        'TRY': {'code': '949', 'name': 'Турецкая лира', 'multiplier': 100},
        'TMT': {'code': '934', 'name': 'Новый туркменский манат', 'multiplier': 100},
        'UZS': {'code': '860', 'name': 'Узбекский сум', 'multiplier': 1000000},
        'UAH': {'code': '980', 'name': 'Гривна', 'multiplier': 1000},
        'CZK': {'code': '203', 'name': 'Чешская крона', 'multiplier': 1000},
        'SEK': {'code': '752', 'name': 'Шведская крона', 'multiplier': 1000},
        'CHF': {'code': '756', 'name': 'Швейцарский франк', 'multiplier': 100},
        'ZAR': {'code': '710', 'name': 'Рэнд', 'multiplier': 1000},
        'KRW': {'code': '410', 'name': 'Вона', 'multiplier': 100000},
        'JPY': {'code': '392', 'name': 'Иена', 'multiplier': 10000}
    }

    def __init__(self, year, broker_as_income=False, only_dividends=False):
        if year not in form_3nfl_template:
            raise ValueError(self.tr("3-NDFL form isn't supoorted for year: ") + f"{year}")
        self._only_dividends = only_dividends
        self._broker_as_income = broker_as_income
        self._year = year
        self.currency = None
        self.broker_name = ''
        self.broker_iso_country = "000"
        self._tax_form = deepcopy(form_3nfl_template[year])
        self.stored_data = {
            "Дивиденды": {"dividend": self.append_dividend},
            "Акции": {"trade": self.append_stock_trade},
            "Облигации": {"bond_trade": self.append_stock_trade, "bond_interest": self.append_bond_interest},
            "ПФИ": {"trade": self.append_derivative_trade},
            "Проценты": {"interest": self.append_other_income}
        }

        self._records = []
        self._sections = {}
        self._footer_len = 0  # if file ends with _footer_len 0x00 bytes

    def tr(self, text):
        return QApplication.translate("DLSG", text)

    # takes tax_report, generated by TaxesRus.prepare_tax_report(), and puts its data into tax form
    # Parameters contain auxiliary params of report
    def update_taxes(self, tax_report, parameters):
        try:
            self.currency = self.currencies[parameters['currency']]
        except KeyError:
            return logging.error(self.tr("Currency is not supported for 3-NDFL ") + parameters['currency'])
        self.broker_name = parameters['broker_name']
        self.broker_iso_country = parameters['broker_iso_country']
        for section in tax_report:
            if section in self.stored_data:
                for template in self.stored_data[section]:
                    for item in tax_report[section]:
                        if item['report_template'] == template:
                            self.stored_data[section][template](item)

    # Save tax form in file format of russian tax software "Декларация" with given filename
    def save(self, filename):
        raw_data = self._tax_form['header']
        for section in self._tax_form['sections']:
            raw_data += self.convert_section(section, self._tax_form['sections'][section])
        with open(filename, "w", encoding='cp1251') as taxes:
            taxes.write(raw_data)

    # Converts one section of self._tax_form into text format of dcX-file
    # Returns a line of text with converted data
    def convert_section(self, section_name, section_data):
        data = self.convert_item(section_name)
        if type(section_data) == tuple:
            for item in section_data:
                data += self.convert_item(item)
        elif type(section_data) == dict:
            # Here is a subsection - need to put length and then process elements
            subitems_number = str(len(section_data))
            data += "{:04d}{}".format(len(subitems_number), subitems_number)
            for sub_item in section_data:
                data += self.convert_section(sub_item, section_data[sub_item])
        return data

    # Converts one field of self._tax_form into text format of dcX-file
    # returns converted value
    def convert_item(self, value):
        if type(value) == str:
            prepared_value = value
        elif type(value) == int or type(value) == float:
            prepared_value = str(value)
        elif type(value) == datetime:
            prepared_value = str((value.date() - date(1899, 12, 30)).days)
        else:
            raise TypeError(f"Unsupported type: {type(value)}")
        data = "{:04d}{}".format(len(prepared_value), prepared_value)
        return data

    def currency_rates_record(self, rate):
        currency_record = (0,  # Auto currency rate
                           self.currency['code'],
                           rate * self.currency['multiplier'],  # Currency rate for income
                           self.currency['multiplier'],  # Currency rate multiplier for income
                           rate * self.currency['multiplier'],  # Currency rate for tax
                           self.currency['multiplier'],  # Currency rate multiplier for tax
                           self.currency['name'])
        return currency_record

    def append_dividend(self, dividend):
        if dividend["country_iso"] == '000':
            logging.error(self.tr(
                "Account country is not set for asset, dividend isn't include in 3-NDFL ") + f"'{dividend['symbol']}'")
            return
        if self._broker_as_income:
            income_source = self.broker_name
            income_iso_country = self.broker_iso_country
        else:
            income_source = f"Дивиденд от {dividend['symbol']} ({dividend['full_name']})"
            income_iso_country = dividend["country_iso"]
        if self._year == 2020:
            income = (14, '1010', 'Дивиденды', income_source, income_iso_country)
        else:
            income = (0, '1010', 'Дивиденды', income_source, income_iso_country, self.broker_iso_country)
        income += (datetime.utcfromtimestamp(dividend['payment_date']),  # Income date
                   datetime.utcfromtimestamp(dividend['payment_date']))  # Tax payment date
        income += self.currency_rates_record(dividend['rate'])
        income += (dividend['amount'], dividend['amount_rub'], dividend['tax'], dividend['tax_rub'])
        if self._year == 2020:
            income += ('0', 0, 0, 0, '', 0)
            items_number = len(self._tax_form['sections']['@DeclForeign'])
            next_label = "@CurrencyIncome{:03d}".format(items_number)
        else:
            income += ('0', 0, 0, 0, 0, '', 0)
            items_number = len(self._tax_form['sections']['@DeclForeign'])
            next_label = "@CurrencyIncome{:04d}".format(items_number)
        self._tax_form['sections']['@DeclForeign'][next_label] = income

    def append_stock_trade(self, trade):
        if trade['qty'] < 0:  # short position - swap close/open dates/rates
            trade['cs_date'] = trade['os_date']
            trade['cs_rate'] = trade['os_rate']
        if self._broker_as_income:
            income_source = self.broker_name
        else:
            income_source = f"Доход от сделки с {trade['symbol']} ({trade['isin']})"
        income_iso_country = self.broker_iso_country
        if self._year == 2020:
            income = (13, '1530', '(01)Доходы от реализации ЦБ (обращ-ся на орг. рынке ЦБ)',
                      income_source, income_iso_country)
        else:
            income = (0, '1530', '(01)Доходы от реализации ЦБ (обращ-ся на орг. рынке ЦБ)',
                      income_source, income_iso_country, income_iso_country)
        income += (datetime.utcfromtimestamp(trade['cs_date']),  # Income date
                   datetime.utcfromtimestamp(trade['cs_date']))  # Tax payment date
        income += self.currency_rates_record(trade['cs_rate'])
        income += (trade['income'], trade['income_rub'], 0, 0, '201', trade['spending_rub'])
        if self._year == 2020:
            income += (0, 0, '', 0)
            items_number = len(self._tax_form['sections']['@DeclForeign'])
            next_label = "@CurrencyIncome{:03d}".format(items_number)
        else:
            income += (0, 0, 0, '', 0)
            items_number = len(self._tax_form['sections']['@DeclForeign'])
            next_label = "@CurrencyIncome{:04d}".format(items_number)
        self._tax_form['sections']['@DeclForeign'][next_label] = income

    def append_bond_interest(self, interest):
        if self._broker_as_income:
            income_source = self.broker_name
        else:
            income_source = f"Купонный доход от {interest['symbol']} ({interest['isin']})"
        income_iso_country = self.broker_iso_country
        if self._year == 2020:
            income = (13, '1530', '(01)Доходы от реализации ЦБ (обращ-ся на орг. рынке ЦБ)',
                      income_source, income_iso_country)
        else:
            income = (0, '1530', '(01)Доходы от реализации ЦБ (обращ-ся на орг. рынке ЦБ)',
                      income_source, income_iso_country, income_iso_country)
        income += (datetime.utcfromtimestamp(interest['o_date']),  # Income date
                   datetime.utcfromtimestamp(interest['o_date']))  # Tax payment date
        income += self.currency_rates_record(interest['rate'])
        income += (interest['interest'], interest['interest_rub'])
        if self._year == 2020:
            income += (0, 0, '0', 0, 0, 0, '', 0)
            items_number = len(self._tax_form['sections']['@DeclForeign'])
            next_label = "@CurrencyIncome{:03d}".format(items_number)
        else:
            income += (0, 0, '0', 0, 0, 0, 0, '', 0)
            items_number = len(self._tax_form['sections']['@DeclForeign'])
            next_label = "@CurrencyIncome{:04d}".format(items_number)
        self._tax_form['sections']['@DeclForeign'][next_label] = income

    def append_derivative_trade(self, trade):
        if trade['qty'] < 0:  # short position - swap close/open dates/rates
            trade['cs_date'] = trade['os_date']
            trade['cs_rate'] = trade['os_rate']
        if self._broker_as_income:
            income_source = self.broker_name
        else:
            income_source = f"Доход от сделки с {trade['symbol']}"
        income_iso_country = self.broker_iso_country
        if self._year == 2020:
            income = (13, '1532',
                      '(06)Доходы по оп-циям с ПФИ (обращ-ся на орг. рынке ЦБ), баз. ак. по которым являются ЦБ',
                      income_source, income_iso_country)
        else:
            income = (0, '1532',
                      '(06)Доходы по оп-циям с ПФИ (обращ-ся на орг. рынке ЦБ), баз. ак. по которым являются ЦБ',
                      income_source, income_iso_country, income_iso_country)
        income += (datetime.utcfromtimestamp(trade['cs_date']),  # Income date
                   datetime.utcfromtimestamp(trade['cs_date']))  # Tax payment date
        income += self.currency_rates_record(trade['cs_rate'])
        income += (trade['income'], trade['income_rub'], 0, 0, '206', trade['spending_rub'])
        if self._year == 2020:
            income += (0, 0, '', 0)
            items_number = len(self._tax_form['sections']['@DeclForeign'])
            next_label = "@CurrencyIncome{:03d}".format(items_number)
        else:
            income += (0, 0, 0, '', 0)
            items_number = len(self._tax_form['sections']['@DeclForeign'])
            next_label = "@CurrencyIncome{:04d}".format(items_number)
        self._tax_form['sections']['@DeclForeign'][next_label] = income

    def append_other_income(self, payment):
        income_source = self.broker_name
        income_iso_country = self.broker_iso_country
        if self._year == 2020:
            income = (13, '4800', 'Иные доходы', income_source, income_iso_country)
        else:
            income = (0, '4800', 'Иные доходы', income_source, income_iso_country, income_iso_country)
        income += (datetime.utcfromtimestamp(payment['payment_date']),  # Income date
                   datetime.utcfromtimestamp(payment['payment_date']))  # Tax payment date
        income += self.currency_rates_record(payment['rate'])
        income += (payment['amount'], payment['amount_rub'])
        if self._year == 2020:
            income += (0, 0, '0', 0, 0, 0, '', 0)
            items_number = len(self._tax_form['sections']['@DeclForeign'])
            next_label = "@CurrencyIncome{:03d}".format(items_number)
        else:
            income += (0, 0, '0', 0, 0, 0, 0, '', 0)
            items_number = len(self._tax_form['sections']['@DeclForeign'])
            next_label = "@CurrencyIncome{:04d}".format(items_number)
        self._tax_form['sections']['@DeclForeign'][next_label] = income
